from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Union

from cerberus import Validator

from forge_template.user_interface import UserInterface
from forge_template.util.dict_util import merge_dictionaries
from forge_template.util.file_util import (
    dump_toml_to_stdout,
    dump_yaml_to_stdout,
    load_toml,
    load_yaml,
    save_toml,
    save_yaml,
)


class FileType(Enum):
    TOML = "toml"
    YAML = "yaml"


class ConfigGenerator:
    def __init__(self, file_type: FileType = FileType.YAML):
        self.ui = UserInterface()
        self.file_type = file_type
        self.config_schema: Dict[str, Any] = {}
        self.output_path: Union[Path, None] = None
        self.config: Dict[str, Any] = {}
        self.key = ""
        self.post_transforms: List[Callable[[str, str], str]] = []
        self.project_name = ""

    def set_schema_info(
        self, *, output_path: Path, schema_path: Path, key: str, post_transforms: List[Callable[[str, str], str]]
    ):
        self.config_schema = load_yaml(schema_path)
        self.output_path = output_path
        self.key = key
        self.post_transforms = post_transforms
        self.config = {}

        if self.file_type == FileType.YAML:
            self.config = load_yaml(output_path) if output_path.exists() else {}
        elif self.file_type == FileType.TOML:
            self.config = load_toml(output_path) if output_path.exists() else {}

    def generate(self, is_custom: bool = False) -> None:
        self.config = self._generate_config(self.key, self.config_schema[self.key], is_custom=is_custom)
        for transform in self.post_transforms:
            self._perform_post_transform(transform, self.config[self.key])

    def preview(self) -> None:
        print()
        self.ui.print_ok(f"Preview: {Path(str(self.output_path)).name}")
        if self.file_type == FileType.YAML:
            dump_yaml_to_stdout(self.config)
        elif self.file_type == FileType.TOML:
            dump_toml_to_stdout(self.config)

    def save(self) -> None:
        print()
        if self.file_type == FileType.YAML:
            save_yaml(Path(str(self.output_path)), self.config)
            self.ui.print_ok(f"Saved {self.output_path}")
        elif self.file_type == FileType.TOML:
            save_toml(Path(str(self.output_path)), self.config)
            self.ui.print_ok(f"Saved {self.output_path}")

    def _generate_config(self, key: str, schema: Dict[str, Any], is_custom: bool = False) -> Dict[str, Any]:
        config = {}
        if type(schema) is dict:
            if schema["type"] == "dict":
                config[key] = merge_dictionaries(
                    [self._generate_config(k, s, is_custom) for k, s in schema["schema"].items()]
                )

            elif schema["type"] == "list" and schema["schema"]["type"] == "dict":
                return {key: self._get_input_list_of_dict(schema)}
            else:
                default = self._get_default_value(schema)
                if not is_custom and default:
                    return {key: default}
                else:
                    transform = self._get_transformer(schema)
                    value = transform(self._get_input(schema, key, default, transform))
                    if key == "project_name":
                        self.project_name = value
                    return {key: value}

        return config

    def _get_input_list_of_dict(self, schema):
        return self.ui.get_input_list_of_dicts(
            prompt=schema["meta"]["prompt"],
            validators=self._get_validators_for_list_of_dicts(schema["schema"]["schema"]),
            keys=list(dict(schema["schema"]["schema"]).keys()),
        )

    def _get_input(self, schema: Dict, key: str, default: str, transform: Callable[[str], Any]) -> str:
        return self.ui.get_input(
            prompt=schema["meta"]["prompt"], default=default, validate=self._get_validator(key, schema, transform)
        )

    def _get_validators_for_list_of_dicts(self, schema: Dict[str, Any]) -> Dict[str, Callable[[str], bool]]:
        validators = {}
        for key in schema:
            validators[key] = self._get_validator(key, schema[key])

        return validators

    def _get_validator(
        self, key: str, schema: Dict[str, Any], transform: Callable[[str], Any] = None
    ) -> Callable[[str], bool]:
        v = Validator({key: schema})
        ui = self.ui

        def validator(input_value: str) -> bool:
            if transform:
                input_value = transform(input_value)
            is_valid = v.validate({key: input_value})
            if not is_valid:
                ui.print_warning(f"Error message: {v.errors}")
                if "meta" in schema and "help" in schema["meta"]:
                    ui.print_warning(f"Explanation: {schema['meta']['help']}")
            return is_valid

        return validator

    @staticmethod
    def _get_transformer(schema: Dict[str, Any]) -> Callable[[str], Any]:
        def transform_list(input_value: str) -> List[str]:
            if input_value:
                return list(map(lambda x: x.strip(), input_value.split(",")))
            return []

        def base_transform(input_value: str) -> str:
            if input_value:
                return input_value.strip()
            return ""

        if schema["type"] == "list":
            return transform_list
        else:
            return base_transform

    @staticmethod
    def _get_default_value(schema: Dict[str, Any]) -> str:
        return schema["default"] if "default" in schema else ""

    def _perform_post_transform(self, transform: Callable[[str, str], str], config: Dict) -> None:
        for k, v in config.items():
            if type(v) is dict:
                self._perform_post_transform(transform, v)
            elif type(v) is str:
                config[k] = transform(v, self.project_name)



import logging
import sys

import click

from forge_template.actions import delete, deploy, generate
from forge_template.constants import Tools
from forge_template.sa_actions import extractor_workflow, generate_data_replication
from forge_template.user_interface import UserInterface

LOGGING_FILE = "logs.log"


GENERATE_OPTIONS = {tool.name.capitalize(): (lambda tool=tool: generate(tool)) for tool in Tools.get_values()}
DEPLOY_OPTIONS = {tool.name.capitalize(): (lambda tool=tool: deploy(tool)) for tool in Tools.get_values()}
DELETE_OPTIONS = {tool.name.capitalize(): (lambda tool=tool: delete(tool)) for tool in Tools.get_values()}
DATA_REPLICATION_OPTIONS = {
    "Set up extractor workflow": lambda: extractor_workflow(),
    "Generate custom data replication": lambda: generate_data_replication(),
}


class ForgeUI(UserInterface):
    def __init__(self):
        main_options = {
            "Generate config": lambda: self.show_menu("Which tool do you want to generate?", GENERATE_OPTIONS),
            "Deploy config": lambda: self.show_menu("Which tool do you want to deploy?", DEPLOY_OPTIONS),
            "Delete resources": lambda: self.show_menu("Which tool do you want to delete?", DELETE_OPTIONS),
            "Extractor workflow & data replication": lambda: self.show_menu(
                "What do you want to do?", DATA_REPLICATION_OPTIONS
            ),
        }

        self.print_title("Forge Template")
        logging.info("If you have any questions that are not answered by the readme, please contact Team Forge\n")
        self.show_menu("What do you want to do?", main_options, is_top_level=True)
        self.print_title("Good bye!")


@click.group(invoke_without_command=True)
def cli():
    _configure_logger()
    ForgeUI()


def _configure_logger():
    logging.basicConfig(filename=LOGGING_FILE, level=logging.INFO)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    logging.getLogger("requests").setLevel(logging.WARNING)
    logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))



from pathlib import Path
from typing import List, Union

import forge_template.paths as file_paths
import forge_template.util.file_util as file_util
import forge_template.util.log_util as log_util
from forge_template.handler.handler import BaseHandler
from forge_template.tool_info import ToolInfo

GITHUB_ACTION_HANDLER_NAME = "github_actions"


class GithubActionsHandler(BaseHandler):
    def __init__(self, tool_info: ToolInfo) -> None:
        self._tool = tool_info.name

        self._output_paths = [file_paths.SCRIPT_OUTPUT_DIR, file_paths.WORKFLOW_OUTPUT_DIR]
        self._output_paths_to_create: List[Path] = []
        self._output_paths_created: List[Path] = []

        self._files_to_add: List[Union[Path, None]] = [
            tool_info.github_actions_script_path,
            tool_info.github_actions_template_path,
        ]  # type: ignore
        self._files_added: List[Path] = []

        super().__init__(config={}, tool_info=tool_info)

    def create_preview(self) -> None:
        self._output_paths_to_create = [path for path in self._output_paths if not path.exists()]

    def print_preview(self) -> None:
        log_util.print_resource_to_add(list(map(str, self._output_paths_to_create)), "Directory")
        log_util.print_resource_to_add(list(map(str, map(lambda x: Path(str(x)).name, self._files_to_add))), "File")

    def setup(self) -> None:
        self._create_paths(self._output_paths_to_create)
        for path, folder in zip(self._files_to_add, self._output_paths):
            self._copy_file(Path(str(path)), folder)

    def rollback(self) -> None:
        if self._files_added:
            log_util.print_rollback("Files")
            self._delete_paths(self._files_added)

    def delete_all_resources(self):
        self._delete_paths(self._output_paths)

    def _copy_file(self, path: Path, destination_folder: Path) -> None:
        file_util.copy_file(path, destination_folder)
        self._files_added.append(destination_folder / path.name)
        log_util.print_success_added(destination_folder, "Directory", path.name, "File")

    @staticmethod
    def _create_paths(paths: List[Path]) -> None:
        for path in paths:
            file_util.create_directory(path, exist_ok=True, parents=True)
            log_util.print_success_created(path.name, "Directory")

    @staticmethod
    def _delete_paths(paths: List[Path]) -> None:
        for path in paths:
            if path.exists():
                file_util.delete_path(path)
                log_util.print_success_deleted(path, "Directory/File")



# Databricks notebook source
# MAGIC %md # Setup base model (database and views)
# MAGIC 
# MAGIC ### Prerequisites
# MAGIC Python 3.7+
# MAGIC 
# MAGIC ### About this notebook
# MAGIC This notebook create a database with views. There is one view for each of the following CDF resource types: assets, events, datapoints, timeseries, files, 3dmodels.
# MAGIC 
# MAGIC - **Cmd 2**: The setup-environment notebook is run to setup the development environment (installing commonly used libraries and reading a configuration file). The parameters from the configuration file is available as a Python dictionary named _config_.

# COMMAND ----------

# MAGIC %run ../setup/setup-environment.py

# COMMAND ----------

# Will create tables and database if True. If False the database will be deleted (if it exists)
CREATE = True 

# COMMAND ----------

# Create database:
if CREATE:
  spark.sql(
      """
      CREATE DATABASE IF NOT EXISTS {database_name} 
      COMMENT '{project_description}'
      """.format(
          **config
      )
  )

# COMMAND ----------

# Grant permissions:
if CREATE:
  spark.sql("GRANT ALL PRIVILEGES ON DATABASE {database_name} TO `{group_name}`".format(**config))

# COMMAND ----------

# Create table with assets
if CREATE:
  for cdf_type in ["assets", "events"]:
    spark.sql("DROP TABLE IF EXISTS {database_name}.api_{cdf_type}".format(cdf_type=cdf_type, **config))
    spark.sql(
        """
    CREATE TABLE IF NOT EXISTS {database_name}.api_{cdf_type} USING cognite.spark.v1
    OPTIONS (type '{cdf_type}',
             apiKey '{api_key}',
             partitions '80')
    COMMENT 'View of {cdf_type} in CDF from base model'
    """.format(
            cdf_type=cdf_type, **config
        )
    )

# COMMAND ----------

if CREATE:
  for cdf_type in ["datapoints", "timeseries", "files", "3dmodels"]:
    spark.sql("DROP TABLE IF EXISTS {database_name}.api_{cdf_type}".format(cdf_type=cdf_type, **config))
    spark.sql(
        """
    CREATE TABLE IF NOT EXISTS {database_name}.api_{cdf_type} USING cognite.spark.v1
    OPTIONS (type '{cdf_type}',
             apiKey '{api_key}')
    COMMENT 'View of {cdf_type} in CDF from base model'
    """.format(
            cdf_type=cdf_type, **config
        )
    )

# COMMAND ----------

if not CREATE:
  spark.sql("DROP DATABASE IF EXISTS {database_name} CASCADE".format(**config))

# COMMAND ----------

# Display database table metadata
if CREATE:
  display(spark.catalog.listTables(config["database_name"]))

# COMMAND ----------

# Display schemas for database tables
if CREATE:
  for t in ["api_assets", "api_datapoints", "api_events", "api_timeseries", "api_files", "api_3dmodels"]:
      print(">>", t)
      spark.table("%s.%s" % (config["database_name"], t)).printSchema()



from pathlib import Path

ROOT_DIR = Path(__file__).parent.parent
PACKAGE_NAME = "forge_template"

CODE_DIR = ROOT_DIR / PACKAGE_NAME
TEMPLATE_DIR = CODE_DIR / "templates"

# Schema paths
SCHEMA_DIR = CODE_DIR / "schema"
YAML_CONFIG_SCHEMA_PATH = SCHEMA_DIR / "config_schema.yaml"
YAML_SECRETS_SCHEMA_PATH = SCHEMA_DIR / "secrets_schema.yaml"
JOB_CONFIG_SCHEMA_PATH = SCHEMA_DIR / "job_config_schema.yaml"

# Config paths
YAML_CONFIG_PATH = Path("config.yaml")
YAML_SECRETS_PATH = Path("secrets.yaml")

# Databricks paths
DATABRICKS_TEMPLATE_DIR = TEMPLATE_DIR / "databricks"
DATABRICKS_OUTPUT_DIR = Path("databricks")

# Power BI paths
POWERBI_TEMPLATE_DIR = TEMPLATE_DIR / "powerbi"
POWERBI_OUTPUT_DIR = Path("powerbi")
POWERBI_BASE_DATASET_PATH = POWERBI_OUTPUT_DIR / "datasets" / "base_dataset.pbix"
POWERBI_BASE_REPORT_PATH = POWERBI_OUTPUT_DIR / "reports" / "base_report.pbix"

# SA Replication paths
# Local paths
DATA_REPLICATOR_OUTPUT_DIR = Path("data_replicator")
EW_OUTPUT_DIR = DATA_REPLICATOR_OUTPUT_DIR / "extractor_workflow"
CR_OUTPUT_DIR = DATA_REPLICATOR_OUTPUT_DIR / "custom_replicator"
DATA_REPLICATOR_TEMPLATE_DIR = TEMPLATE_DIR / "data_replicator"
JOB_CONFIG_TEMPLATE = DATA_REPLICATOR_TEMPLATE_DIR / "extractor_workflow" / "job_config_file_template.yaml"
EW_SCHEDULE_TEMPLATE = DATA_REPLICATOR_TEMPLATE_DIR / "extractor_workflow" / "schedule_config_file_template.txt"
CR_SCHEDULE_TEMPLATE = DATA_REPLICATOR_TEMPLATE_DIR / "custom_replicator" / "schedule_config_file_template.txt"
# GCP paths
EW_GCP_CONFIG_PATH = Path("extractor_workflow")
EW_GCP_SCHEDULE_PATH = Path("dags", "extractor_workflow")
CR_GCP_CONFIG_PATH = Path("custom_replicator")
CR_GCP_SCHEDULE_PATH = Path("dags", "custom_replicator")

# Github Actions paths
SCRIPT_TEMPLATE_DIR = TEMPLATE_DIR / "infra_scripts"
POWERBI_SCRIPT_PATH = SCRIPT_TEMPLATE_DIR / "deploy_powerbi.py"
DATABRICKS_SCRIPT_TEMPLATE_PATH = SCRIPT_TEMPLATE_DIR / "deploy_databricks.py"
SCRIPT_OUTPUT_DIR = Path("pipeline_scripts")

WORKFLOW_TEMPLATE_DIR = TEMPLATE_DIR / "github_workflows"
POWERBI_WORKFLOW_TEMPLATE_PATH = WORKFLOW_TEMPLATE_DIR / "deploy_powerbi.yaml"
DATABRICKS_WORKFLOW_TEMPLATE_PATH = WORKFLOW_TEMPLATE_DIR / "deploy_databricks.yaml"
WORKFLOW_OUTPUT_DIR = Path(".github") / "workflows"



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Callable, List, Type

from forge_template.config_generator import FileType

if TYPE_CHECKING:
    from forge_template.handler.handler import BaseHandler


class SchemaInfo:
    def __init__(self, schema_path: Path, output_path: Path, post_transforms: List[Callable[[str, str], str]] = None):
        self.schema_path = schema_path
        self.output_path = output_path
        self.post_transforms = post_transforms


class ToolInfo:
    def __init__(
        self,
        name: str,
        handler: Type[BaseHandler],
        schema_info: List[SchemaInfo],
        github_actions_script_path: Path = None,
        github_actions_template_path: Path = None,
        assert_match: bool = False,
        assert_scope_match: bool = False,
        file_type: FileType = FileType.YAML,
        generate_all: bool = True,
    ):
        self.name = name
        self.handler = handler
        self.schema_info = schema_info
        self.github_actions_script_path = github_actions_script_path
        self.github_actions_template_path = github_actions_template_path
        self.assert_match = assert_match
        self.assert_scope_match = assert_scope_match
        self.file_type = file_type
        self.generate_all = generate_all



from collections.abc import Mapping
from pathlib import Path
from typing import Any, Dict, List


def assert_same_keys(configs: Dict[Path, Dict[str, Any]]) -> None:
    if len(configs) == 0:
        return
    config_keys = {key: configs[key].keys() for key in configs}

    base_name = list(config_keys.keys())[0]
    base_keys = config_keys.pop(base_name)

    for name, keys in config_keys.items():
        is_valid = base_keys == keys
        diff = {}
        if len(base_keys - keys) != 0:
            diff[f"{name} is missing"] = base_keys - keys
        if len(keys - base_keys) != 0:
            diff[f"{base_name} is missing"] = keys - base_keys

        assert is_valid, f"YAML files don't contain the same services defined. Differences:\n{diff}"


def merge_dictionaries(configs: List[Dict]) -> Dict:
    base = configs[0].copy()
    if len(configs) == 1:
        return base

    for d in configs[1:]:
        merge_two_dictionaries(base, d)

    return base


def merge_two_dictionaries(dict_a: Dict, dict_b: Dict) -> None:
    for k, v in dict_b.items():
        if k in dict_a and isinstance(dict_a[k], dict) and isinstance(dict_b[k], Mapping):
            merge_two_dictionaries(dict_a[k], dict_b[k])
        else:
            dict_a[k] = dict_b[k]


def assert_scope_matches(configs: Dict[Path, Dict[str, Any]]) -> None:
    if len(configs) != 2:
        return
    keys = list(configs.keys())

    key1 = keys[0]
    config1 = configs[key1]
    key2 = keys[1]
    config2 = configs[key2]

    if "databricks" in config1 and "databricks" in config2:
        for stage in ["production", "development"]:
            name_in_config1 = config1["databricks"][stage]["scope"]["name"]
            name_in_config2 = config2["databricks"][stage]["scope"]["name"]
            assert name_in_config1 == name_in_config2, (
                f"Mismatch between scope name in '{key1}' ({name_in_config1}) and " f"'{key2}' ({name_in_config2})."
            )



# import os
# import re
# import shutil
# from typing import Dict, List
#
# from pbiapi import PowerBIAPIClient
#
# import forge_template.util.log_util as log_util
# from forge_template.handler.handler import BaseHandler
# from forge_template.paths import (
#     POWERBI_BASE_DATASET_PATH,
#     POWERBI_BASE_REPORT_PATH,
#     POWERBI_PATH,
#     POWERBI_TEMPLATE_PATH,
# )
# from forge_template.tool_info import ToolInfo
#
#
# class PowerBIHandler(BaseHandler):
#     def __init__(self, config: Dict, tool_info: ToolInfo) -> None:
#         self.name = tool_info.name
#         config = config[self.name]
#         self._client = PowerBIAPIClient(config["tenant_id"], config["application_id"], config["token"])
#
#         self._dev_workspace = config["development"]
#         self._prod_workspace = config["production"]
#
#         self._workspaces_to_create: List[str] = []
#         self._users_to_add_to_workspace: Dict[str, List[str]] = {}
#         self._users_to_update_in_workspace: Dict[str, List[Dict]] = {}
#
#         self._added_workspaces: List[str] = []
#         self._added_users_to_workspace: Dict[str, List[str]] = {}
#
#         super().__init__(config=config, tool_info=tool_info)
#
#     def create_preview(self) -> None:
#         self.check_workspaces_to_add([self._dev_workspace, self._prod_workspace])
#
#     def print_preview(self) -> None:
#         if not self._workspaces_to_create and not self._users_to_add_to_workspace:
#             log_util.print_message("No changes to deploy")
#         log_util.print_resource_to_add(self._workspaces_to_create, "Workspace")
#         log_util.print_resources_to_update(self._users_to_add_to_workspace, "Principal", "Workspace")
#
#     def setup(self) -> None:
#         self.create_workspaces(self._workspaces_to_create)
#         self.add_users_to_workspace(self._users_to_add_to_workspace)
#
#     def rollback(self) -> None:
#         if self._added_workspaces:
#             log_util.print_rollback("workspaces")
#             self.__remove_workspaces(self._added_workspaces)
#
#     def delete_all_resources(self):
#         self.__remove_workspaces([self._dev_workspace["name"], self._prod_workspace["name"]])
#
#     def check_workspaces_to_add(self, workspaces: List) -> None:
#         """
#         Checks which PowerBI workspaces to add, as well as which users to add to the workspace.
#
#         Args:
#             workspaces: List of workspace definitions. See config.yaml for example.
#         """
#         existing_workspaces = self.__get_existing_workspaces()
#         for workspace in workspaces:
#             workspace_name = workspace["name"]
#             if workspace_name not in existing_workspaces.keys():
#                 self._workspaces_to_create.append(workspace_name)
#                 self._users_to_add_to_workspace[workspace_name] = workspace["principals"]
#
#     def __get_existing_workspaces(self) -> Dict:
#         response = self._client.get_workspaces()
#         return {w["name"]: w for w in response}
#
#     def __remove_workspaces(self, workspace_names: List) -> None:
#         for workspace_name in workspace_names:
#             self._client.delete_workspace(workspace_name)
#             log_util.print_success_deleted(workspace_name, "Workspace")
#
#     def create_workspaces(self, workspace_names: List) -> None:
#         """
#         Creates workspaces in PowerBI
#
#         Args:
#             workspace_names: Names of workspaces to create
#         """
#         self.__copy_folders_locally()
#         for name in workspace_names:
#             self.create_workspace(name)
#
#     def create_workspace(self, workspace_name: str) -> None:
#         """
#         Creates a workspace in PowerBI and uploads a dataset to that workspace.
#
#         Args:
#             workspace_name: Name of workspace to create
#         """
#         self._client.create_workspace(workspace_name)
#         self._added_workspaces.append(workspace_name)
#         self.__upload_files_to_workspace(workspace_name)
#         self.__rebind_report_to_dataset(workspace_name)
#         log_util.print_success_created(workspace_name, "Workspace")
#
#     def add_users_to_workspace(self, workspace_name_to_users: Dict[str, List[str]]) -> None:
#         """
#         Add a users to an existing PowerBI workspace.
#
#         Args:
#             workspace_name_to_users: Workspace name to list of users where each user is a dictionary with keys:
#                 ``"identifier"``
#                 Id of user to add (email, group id or app id) (str)
#
#                 ``"access_right"``
#                 Permission level to give to the user (Admin, Member, Contributor) (str)
#         """
#         for workspace_name, users in workspace_name_to_users.items():
#             for user in users:
#                 self.add_user_to_workspace(user, workspace_name)
#
#     def add_user_to_workspace(self, user_id: str, workspace_name: str, access_right: str = "Admin") -> None:
#         """
#         Add a user to an existing workspace and set the permissions of that user
#
#         Args:
#             user_id: Id of user. A user can be a person, group or Power App.
#             workspace_name: Name of workspace where the user will be added
#             access_right: Permission level to give to the user
#         """
#         user = {"identifier": user_id, "groupUserAccessRight": access_right}
#         if self.__assert_is_email(user_id):
#             user["UserEmailAddress"] = user_id
#         self._client.add_user_to_workspace(workspace_name, user)
#         if workspace_name not in self._added_users_to_workspace.keys():
#             self._added_users_to_workspace[workspace_name] = []
#         self._added_users_to_workspace[workspace_name].append(user_id)
#         log_util.print_success_added(workspace_name, "Workspace", user_id, "Principal")
#
#     def __upload_files_to_workspace(self, workspace_name: str) -> None:
#         pbix_paths = [POWERBI_BASE_DATASET_PATH, POWERBI_BASE_REPORT_PATH]
#         for file_path in pbix_paths:
#             display_name = self.get_display_name_from_pbix_path(file_path)
#             skip_report = "dataset" in display_name
#             self.__upload_file_to_workspace(workspace_name, file_path, display_name, skip_report)
#
#     def __upload_file_to_workspace(
#         self, workspace_name: str, local_path: str, display_name: str, skip_report: bool = False
#     ) -> None:
#         self._client.import_file_into_workspace(workspace_name, skip_report, local_path, display_name)
#         log_util.print_success_added(workspace_name, "Workspace", display_name, "PBIX file")
#
#     def __rebind_report_to_dataset(self, workspace_name: str) -> None:
#         dataset_name = self.get_display_name_from_pbix_path(POWERBI_BASE_DATASET_PATH)
#         report_name = self.get_display_name_from_pbix_path(POWERBI_BASE_REPORT_PATH)
#
#         self._client.rebind_report_in_workspace(workspace_name, dataset_name, report_name)
#         self._client.delete_dataset(workspace_name, report_name)
#
#     @staticmethod
#     def __assert_is_email(s: str) -> bool:
#         regex = re.compile(r"^\w+([.-]?\w+)*@\w+([.-]?\w+)*(\.\w{2,3})+$")
#         return re.search(regex, s) is not None
#
#     @staticmethod
#     def __copy_folders_locally() -> None:
#         if not os.path.exists(POWERBI_PATH):
#             shutil.copytree(POWERBI_TEMPLATE_PATH, POWERBI_PATH)
#
#     @staticmethod
#     def get_display_name_from_pbix_path(path: str) -> str:
#         return os.path.splitext(os.path.basename(path))[0]



import readline  # noqa: allows for arrow up/down
import sys
from typing import Callable, Dict, List, Optional

import colorama
from click import confirm
from pyfiglet import Figlet
from termcolor import colored, cprint

from forge_template.util.log_util import print_goodbye


class UserInterface:
    RETURN = "b"
    EXIT = "e"

    colorama.init()

    @staticmethod
    def print_title(title):
        cprint(Figlet(font="slant").renderText(title), "white")

    @staticmethod
    def print_failure(msg):
        cprint(msg, "red")

    @staticmethod
    def print_warning(msg):
        cprint(msg, "yellow")

    @staticmethod
    def print_ok(msg):
        cprint(msg, "green")

    @staticmethod
    def print_in_progress(msg):
        cprint(msg, "blue")

    def get_input(
        self,
        prompt: str,
        default: str = "",
        choices: Optional[List[str]] = None,
        validate: Optional[Callable[[str], bool]] = None,
    ) -> str:
        _prompt = self.format_prompt(prompt, choices, default)
        value = input(_prompt) or default
        if choices and value not in choices:
            self.print_failure("Must be one of {}".format(choices))
            return self.get_input(prompt, default, choices, validate)
        if validate and validate(value) is False:
            return self.get_input(prompt, default=default, choices=choices, validate=validate)
        return value

    def show_menu(self, prompt: str, options: Dict[str, Callable], is_top_level: bool = False) -> str:
        choice = ""
        while choice != self.RETURN or is_top_level:
            choice = self._get_menu_selection(prompt, list(options.keys()), is_top_level=is_top_level)
            if choice == self.EXIT:
                print_goodbye()
                sys.exit(0)  # Clean exit, regardless of nested calls
            if choice != self.RETURN or is_top_level:
                choice = options[choice]()
        return choice

    def choose_option(self, prompt, options: List) -> str:
        choice = self._get_menu_selection(prompt, options, is_top_level=True)
        if choice == self.EXIT:
            print_goodbye()
            sys.exit(0)  # Clean exit, regardless of nested calls
        return choice

    def _get_menu_selection(self, prompt, choices: List[str], is_top_level: bool = False) -> str:
        print()
        for i, choice in enumerate(choices):
            print("[{}] {}".format(colored(str(i), "green"), choice))
        print()
        if not is_top_level:
            print("[{}] {}".format(colored(self.RETURN, "green"), "Back"))
        print("[{}] {}".format(colored(self.EXIT, "green"), "Exit"))
        print()
        choice = self.get_input(prompt)

        i_choices = [str(i) for i in range(len(choices))] + ([self.EXIT] if is_top_level else [self.RETURN, self.EXIT])
        while choice not in i_choices:
            self.print_failure("Must be one of {}".format(i_choices))
            choice = self.get_input(prompt)

        if choice in [self.EXIT, self.RETURN]:
            return choice
        return choices[int(choice)]

    def get_input_list_of_dicts(self, prompt: str, validators: Dict[str, Callable[[str], bool]], keys: List[str]):
        result = []
        while confirm(prompt, prompt_suffix="? ", show_default=True, default=False):
            values = {}
            for key in keys:
                values[key] = self.get_input(key.capitalize(), validate=validators[key])

            result.append(values)

        return result

    @staticmethod
    def format_prompt(prompt: str, choices: Optional[List[str]], default: str) -> str:
        _prompt = prompt
        if choices:
            _prompt += colored(" ({})".format("/".join(choices)), "blue")
        if default:
            _prompt += colored(" [{}]".format(default), "cyan")
        return _prompt + ": "



import base64
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import toml
from cerberus import Validator
from ruamel.yaml import YAML, sys

from forge_template.exception.exceptions import ValidationException


def copy_tree_and_replace_placeholders(src_dest_pairs: List[Tuple[Path, Path]], placeholder: str, replacement: str):
    for src, dest in src_dest_pairs:
        dest.parent.mkdir(parents=True, exist_ok=True)
        content = src.open("r").read()
        content = content.replace(placeholder, replacement)
        dest.open("w").write(content)


def get_rel_folder_paths(path: Path) -> List[Path]:
    return [d.relative_to(path) for d in path.rglob("**/*") if d.is_dir()]


def get_base64_encoded_content(path: Path) -> str:
    content = path.open("rb").read()
    return base64.b64encode(content).decode()


def load_yaml(path: Path) -> Dict:
    yaml = YAML(typ="safe")
    dct = yaml.load(path.open("r", encoding="utf-8"))
    assert isinstance(dct, dict), f"YAML malformatted, couldn't load into dict (got {type(dct)})"
    return dct


def load_toml(path: Path) -> Dict:
    dct = toml.load(path.open(encoding="utf-8"))
    assert isinstance(dct, Dict), f"TOML malformatted, coldn't load into dict (got {type(dct)}"
    return dct


def save_yaml(path: Path, content: Dict) -> None:
    yaml = YAML(typ="safe")
    yaml.indent(sequence=4, offset=2)
    yaml.dump(content, path)


def save_toml(path: Path, content: Dict) -> None:
    toml.dump(content, path.open("w"))


def dump_yaml_to_stdout(content: Any) -> None:
    yaml = YAML()
    yaml.indent(sequence=4, offset=2)
    yaml.dump(content, sys.stdout)


def dump_toml_to_stdout(content: Dict) -> None:
    print(toml.dumps(content))


def validate_yaml(schema_path: Path, document: Dict) -> Tuple[bool, Optional[Dict]]:
    schema = load_yaml(schema_path)
    validator = Validator(schema)
    return validator.validate(document), validator.errors or None


def load_and_validate_yaml(path: Path, schema_path: Path) -> Dict:
    loaded_yaml = load_yaml(path)
    is_valid, errors = validate_yaml(schema_path, loaded_yaml)
    if not is_valid:
        raise ValidationException(f"Validation of {path} failed.\n{errors}")
    return loaded_yaml


def read_file(path: Path) -> str:
    return path.open("r").read()


def write_file(path: Path, content: str) -> None:
    path.open("w").write(content)


def create_directory(path: Path, *, exist_ok: bool = False, parents: bool = False) -> None:
    path.mkdir(exist_ok=exist_ok, parents=parents)


def copy_file(path: Path, destination: Path) -> None:
    shutil.copy(str(path), str(destination))


def delete_path(path: Path) -> None:
    if path.is_dir():
        shutil.rmtree(path)
    elif path.is_file():
        path.unlink()


def move_directory(src: Path, dst: Path):
    create_directory(path=dst, parents=True, exist_ok=True)
    src.rename(dst)



# import logging
# from typing import List
#
# from git import Repo
#
# import forge_template.util.log_util as log_utils
#
# FORGE_REMOTE_NAME = "forge-template"
#
#
# def push_content_to_repo(repo: Repo, repo_name: str) -> None:
#     repo.index.add("*")
#     added_files = [key[0] for key in repo.index.entries.keys()]
#     repo.git.commit("-m", "initial commit")
#     repo.git.push("--set-upstream", "origin", "master")
#     for added_file in added_files:
#         log_utils.print_success_added(repo_name, "Repository", added_file, "File")
#
#
# def check_remotes_to_add(repo: Repo, remote_name: str, remotes_to_rename: List, remotes_to_add: List) -> None:
#     origin_exists = is_remote_in_remotes(repo, "origin")
#     if origin_exists:
#         remotes_to_rename.append(["origin", FORGE_REMOTE_NAME])
#         remotes_to_add.append(remote_name)
#
#
# def add_remotes(remotes_to_add: List, url: str, repo: Repo, remotes_added: List) -> None:
#     for remote in remotes_to_add:
#         repo.create_remote(remote, url=url)
#         remotes_added.append(remote)
#         log_utils.print_success_created(remote, "Remote")
#
#
# def rename_remotes(remotes_to_rename: List, repo: Repo, renamed_remotes: List) -> List:
#     for (old_name, new_name) in remotes_to_rename:
#         if is_remote_in_remotes(repo, old_name):
#             repo.remote(old_name).rename(new_name)
#             renamed_remotes.append((new_name, old_name))
#             logging.info(f"Renamed remote {old_name} to {new_name}")
#     return renamed_remotes
#
#
# def remove_remotes(remote_list: List, repo: Repo) -> None:
#     for remote in remote_list:
#         if is_remote_in_remotes(repo, remote):
#             repo.git.remote("rm", remote)
#             log_utils.print_success_deleted(remote, "Remote")
#
#
# def is_remote_in_remotes(repo: Repo, remote_name: str) -> bool:
#     return remote_name in [remote.name for remote in repo.remotes]



import forge_template.paths as paths
from forge_template.config_generator import FileType
from forge_template.handler.databricks_handler import DatabricksHandler
from forge_template.tool_info import SchemaInfo, ToolInfo


class ConstList:
    @classmethod
    def get_attrs(cls):
        return [
            a for a in dir(cls) if (not a.startswith("_") and a not in ("get_attrs", "get_values", "get_attr_to_value"))
        ]

    @classmethod
    def get_values(cls):
        return [getattr(cls, a) for a in cls.get_attrs()]

    @classmethod
    def get_attr_to_value(cls):
        return {a: getattr(cls, a) for a in cls.get_attrs()}


class Tools(ConstList):
    DATABRICKS = ToolInfo(
        name="databricks",
        handler=DatabricksHandler,
        schema_info=[
            SchemaInfo(
                schema_path=paths.YAML_CONFIG_SCHEMA_PATH,
                output_path=paths.YAML_CONFIG_PATH,
                post_transforms=[
                    lambda s, p: s.replace("<PROJECT_NAME>", p),
                    lambda s, p: s.replace("<DB_PROJECT_NAME>", p.replace("-", "_")),
                ],
            ),
            SchemaInfo(
                schema_path=paths.YAML_SECRETS_SCHEMA_PATH,
                output_path=paths.YAML_SECRETS_PATH,
                post_transforms=[lambda s, p: s.replace("<PROJECT_NAME>", p)],
            ),
        ],
        github_actions_script_path=paths.DATABRICKS_SCRIPT_TEMPLATE_PATH,
        github_actions_template_path=paths.DATABRICKS_WORKFLOW_TEMPLATE_PATH,
        assert_match=True,
        assert_scope_match=True,
        file_type=FileType.YAML,
    )
    # POWERBI = ToolInfo(
    #     name="powerbi",
    #     handler=PowerBIHandler,
    #     schema_info=[
    #         SchemaInfo(schema_path=YAML_CONFIG_SCHEMA_PATH, output_path=YAML_CONFIG_PATH),
    #         SchemaInfo(schema_path=YAML_SECRETS_SCHEMA_PATH, output_path=YAML_SECRETS_PATH),
    #     ],
    #     github_actions_script_name=POWERBI_SCRIPT_NAME,
    #     github_actions_template_name=POWERBI_WORKFLOW_TEMPLATE_NAME,
    #     assert_match=True,
    #     assert_scope_match=True,
    # )



import logging
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, NoReturn, Optional

import click
from cognite.client import CogniteClient
from cognite.client.data_classes import DataSet
from google.api_core.exceptions import NotFound
from google.auth.exceptions import DefaultCredentialsError
from google.cloud import storage
from google.cloud.secretmanager import SecretManagerServiceClient
from termcolor import cprint

from forge_template.handler.sa_file_handler import CustomReplicatorFileHandler, Env, ExtractorWorkflowFileHandler
from forge_template.paths import (
    CR_GCP_CONFIG_PATH,
    CR_OUTPUT_DIR,
    DATA_REPLICATOR_OUTPUT_DIR,
    EW_GCP_CONFIG_PATH,
    EW_OUTPUT_DIR,
)
from forge_template.sa_constants import (
    CDF_API_HOST_REGEX,
    DATA_TYPE_OPTIONS,
    DEFAULT_CDF_API_HOST,
    DOC_URL,
    GCP_PROJECT_NAME,
    GS_SERVICE_ACCOUNTS,
    JOB_CONFIGS_BUCKET,
    SCHEDULE_OPTIONS,
)
from forge_template.user_interface import UserInterface
from forge_template.util.log_util import print_goodbye
from forge_template.util.sa_util import get_utc_timestamp_as_str, print_monitor_replication

logger = logging.getLogger(__file__)


@dataclass
class CDFInfo:
    name: str
    key: str
    client: CogniteClient


def generate_data_replication() -> None:
    """
    Kicks off "Generate a custom data replication"
    """
    ui = UserInterface()

    try:
        secret_manager_client = SecretManagerServiceClient()
    except DefaultCredentialsError as e:
        handle_default_credentials_error(e)

    # Request tenant information
    while True:
        tenant_names = get_tenant_names_data_replication()
        tenant_info = get_tenant_info(secret_manager_client, tenant_names, ui)
        if tenant_info:
            break

    sync_path = CR_GCP_CONFIG_PATH / f"{tenant_names[Env.SRC]}-{tenant_names[Env.DST]}"
    sync_local_directory(sync_path)

    # Request replication information
    data_type = ui.choose_option(prompt="Which data type do you want to replicate?", options=DATA_TYPE_OPTIONS)
    schedule = ui.choose_option(prompt="How often should the data be replicated", options=SCHEDULE_OPTIONS)
    cdf_api_hosts = request_cdf_api_hosts()
    job_id = ""
    while not job_id:
        job_id = click.prompt(f"Please choose a name / id for the replication job")
        if does_job_id_exist(
            job_id=job_id, src_tenant=tenant_names[Env.SRC], dst_tenant=tenant_names[Env.DST], data_type=data_type
        ):
            cprint(f'The chosen job id "{job_id}" already exists.', "red")
            if not click.confirm("Do you want to overwrite the existing job?"):
                job_id = ""

    # Generate job config
    job_config_generated = False
    while not job_config_generated:
        file_handler = CustomReplicatorFileHandler(
            data_type=data_type,
            schedule=schedule,
            tenant_names=tenant_names,
            cdf_api_hosts=cdf_api_hosts,
            job_id=job_id,
        )
        cprint("Please specify the replication details:", color="blue")
        file_handler.generate_job_config()
        file_handler.config_generator.preview()
        if click.confirm("Please check your input once again. Are you satisfied with the shown preview?"):
            file_handler.config_generator.save()
            job_config_generated = True
        else:
            job_config_generated = False
    file_handler.generate_schedule()
    file_handler.upload_configs_to_gcp()
    cprint(
        f"Successfully activated replication of {data_type} from {tenant_names[Env.SRC]} to {tenant_names[Env.DST]}",
        "green",
    )
    print_monitor_replication()


def extractor_workflow():
    """
    Kicks off "Set up extractor workflow"
    """
    ui = UserInterface()

    try:
        secret_manager_client = SecretManagerServiceClient()
    except DefaultCredentialsError as e:
        handle_default_credentials_error(e)

    while True:
        tenant_names = get_tenant_names_extractor_workflow(ui)
        base_tenant = tenant_names.pop(Env.BASE)
        tenant_info = get_tenant_info(secret_manager_client, tenant_names, ui)
        if type(tenant_info) is dict:
            break

    sync_path = EW_GCP_CONFIG_PATH / f"{base_tenant}"
    sync_local_directory(sync_path)

    options = {
        "Work on a new extractor": lambda: new_extractor(tenant_info, base_tenant, ui),
        "Update an existing extractor": lambda: update_extractor(tenant_info, base_tenant, ui),
    }

    ui.show_menu("What do you want to do?", options)


def sync_local_directory(directory: Path):
    """
    Syncs the local directory for storing the config files with the remote one

    Parameters:
         directory(str): which directory to sync. Should be either "custom_replicator" or "extractor_workflow"
    """
    try:
        storage_client = storage.Client()
    except DefaultCredentialsError as e:
        handle_default_credentials_error(e)

    cprint(f"Start syncing local with remote repository: {directory}", "blue")
    bucket_name = JOB_CONFIGS_BUCKET
    prefix = directory
    local_base_dir = DATA_REPLICATOR_OUTPUT_DIR

    blobs = storage_client.list_blobs(bucket_name, prefix=str(prefix))
    download_count = 0
    for blob in blobs:
        file_split = blob.name.split("/")
        file_directory = "/".join(file_split[:-1])
        Path(local_base_dir, file_directory).mkdir(parents=True, exist_ok=True)
        blob.download_to_filename(Path(local_base_dir, blob.name))
        download_count += 1

    if download_count:
        cprint(f"Successfully synced {download_count} files from remote repository", "green")
    else:
        cprint(f"Remote repository seems empty!", "yellow")


def new_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface) -> None:
    """
    Kicks off "Work on a new extractor"

    Parameters:
        tenant_info (Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
    """
    options = {
        f"Start work on a new extractor on {tenant_info[Env.DEV].name}": lambda: start_new_extractor(
            tenant_info, base_tenant, ui
        ),
        f"Test new extractor on {tenant_info[Env.TEST].name}": lambda: test_new_extractor(tenant_info, base_tenant, ui),
        f"Deploy new extractor to {tenant_info[Env.PROD].name}": lambda: deploy_new_extractor(
            tenant_info, base_tenant, ui
        ),
    }

    ui.show_menu("What do you want to do?", options)


def update_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface) -> None:
    """
    Kicks off "Update an existing extractor"

    Parameters:
        tenant_info (Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str)
        ui (UserInterface):
    """
    options = {
        f"Start work on updating an existing extractor on "
        f"{tenant_info[Env.DEV].name}": lambda: start_update_extractor(tenant_info, base_tenant, ui),
        f"Test updated extractor on {tenant_info[Env.TEST].name}": lambda: test_update_extractor(
            tenant_info, base_tenant, ui
        ),
        f"Deploy updated extractor to {tenant_info[Env.PROD].name}": lambda: deploy_update_extractor(
            tenant_info, base_tenant, ui
        ),
    }

    ui.show_menu("What do you want to do?", options)


def start_new_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface) -> None:
    """
    Kicks off "Start work on a new extractor on <tenant-dev>"

    Parameters:
        tenant_info (Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
    """
    start_extractor(tenant_info, base_tenant, ui, update=False)


def test_new_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface) -> None:
    """
    Kicks off "Test new extractor on <tenant-test>"

    Parameters:
        tenant_info (Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
    """
    test_extractor(tenant_info, base_tenant, ui, update=False)


def deploy_new_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface) -> None:
    """
    Kicks of "Deploy new extractor on <tenant-prod>"

    Parameters:
        tenant_info (Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
    """
    deploy_extractor(tenant_info, base_tenant, ui, update=False)


def start_update_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface) -> None:
    """
    Kicks off "Start work on updating an existing extractor on <tenant-dev>"

    Parameters:
        tenant_info(Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
    """
    start_extractor(tenant_info, base_tenant, ui, update=True)


def test_update_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant, ui: UserInterface) -> None:
    """
    Kicks off "Test an updated extractor on <tenant-dev>"

    Parameters:
        tenant_info(Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
    """
    test_extractor(tenant_info, base_tenant, ui, update=True)


def deploy_update_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface) -> None:
    """
    Kicks off "Deploy an updated extractor on <tenant-dev>"

    Parameters:
        tenant_info(Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
    """
    deploy_extractor(tenant_info, base_tenant, ui, update=True)


def start_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface, update: bool):
    """
    Guides the user through either
    1) Start working on a new extractor (update: False) or
    2) Start working on updating an existing extractor (update: True)
    Additionally, the necessary data replications are set up

    Parameters:
        tenant_info(Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
        update (bool): True for "update extractor" workflow, False for "new extractor" workflow
    """
    prod_tenant = tenant_info[Env.PROD].name
    dev_tenant = tenant_info[Env.DEV].name

    # Get extractor_id
    if update:
        try:
            extractor_options = get_extractor_names(location=Path(base_tenant, "prod-test", "deployed"))
            extractor_id = ui.choose_option(prompt="Choose extractor to deploy", options=extractor_options)
        except FileNotFoundError:
            cprint("There is no extractor to start updating. Please select another option.", "red")
            return
    else:
        extractor_id = click.prompt("Please enter external id of extractor / data set")

    # Handle information about which data sets to replicate
    while True:
        data_sets = click.prompt(
            f"Enter which data sets to replicate from {prod_tenant} to {dev_tenant} as a comma separated list"
        )
        data_sets = data_sets.replace(" ", "").split(",")
        invalid_data_sets = get_invalid_data_sets(data_sets, tenant_info[Env.PROD])
        if not invalid_data_sets:
            break
        cprint(f"The following data sets do not exist in {prod_tenant}: {invalid_data_sets}", "red")

    invalid_data_sets_dev = get_invalid_data_sets(data_sets, tenant_info[Env.DEV])
    if not does_data_set_exist(extractor_id, tenant_info[Env.DEV]):
        invalid_data_sets_dev.append(extractor_id)

    for data_set in invalid_data_sets_dev:
        create_data_set(data_set, tenant_info[Env.DEV])

    # Set up replication and store files in prod-dev/started
    schedule = ui.choose_option(prompt="How often should the data be replicated", options=SCHEDULE_OPTIONS)
    file_handler = ExtractorWorkflowFileHandler(base_tenant=base_tenant, extractor_id=extractor_id, schedule=schedule)
    file_handler.generate_config_files(
        data_sets_to_replicate=data_sets, src_env=Env.PROD, dst_env=Env.DEV, stage="started"
    )
    file_handler.activate_replication(src_env=Env.PROD, dst_env=Env.DEV, stage="started")

    # For updating existing extractor: move already existing config files
    if update:
        file_handler.move_folder_locally(src_env=Env.PROD, dst_env=Env.TEST, move_from="deployed", move_to="started")
        file_handler.move_folder_on_gcp(src_env=Env.PROD, dst_env=Env.TEST, move_from="deployed", move_to="started")

    # User guidance: remind to deploy new / updated extractor to development environment
    workflow_stage = "updated" if update else "new"
    cprint(f'Successfully activated the replication of {data_sets} from "{prod_tenant}" to "{dev_tenant}".', "green")
    print_monitor_replication()
    cprint(
        f'Now it is time to deploy the {workflow_stage} extractor towards "{dev_tenant}" and populate the data set '
        f'"{extractor_id}".',
        "yellow",
    )
    click.confirm("Hit enter to confirm deployment", show_default=False)

    # Finally, check for rollback of data replications towards dev-tenant
    rollback = check_for_rollback(tenant=dev_tenant, stage="started", update=update, file_handler=file_handler)

    if not rollback:
        s = "an updated" if update else "a new"
        cprint(
            f"Congratulations! You completed all steps to start working on {s} extractor. Now you can continue with "
            "the test step.",
            "green",
        )


def test_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface, update: bool):
    """
    Guides the user through either
    1) Test a new extractor (update: False) or
    2) Test an updated, existing extractor (update: True)


    Parameters:
        tenant_info(Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
        update (bool): True for "update extractor" workflow, False for "new extractor" workflow
    """
    test_tenant = tenant_info[Env.TEST].name

    # Get extractor name
    read_extractor_location = "prod-test" if update else "prod-dev"
    try:
        extractor_options = get_extractor_names(location=Path(base_tenant, read_extractor_location, "started"))
        extractor_id = ui.choose_option(prompt="Choose extractor to test", options=extractor_options)
    except FileNotFoundError:
        cprint("There is no extractor to test. Please select another option.", "red")
        return

    # Generate necessary data sets on test tenant
    if not does_data_set_exist(extractor_id, tenant_info[Env.TEST]):
        create_data_set(extractor_id, tenant_info[Env.TEST])

    # Move files from prod-dev/started -> prod-dev/tested
    file_handler = ExtractorWorkflowFileHandler(base_tenant=base_tenant, extractor_id=extractor_id)
    file_handler.move_folder_locally(src_env=Env.PROD, dst_env=Env.DEV, move_from="started", move_to="tested")
    file_handler.move_folder_on_gcp(src_env=Env.PROD, dst_env=Env.DEV, move_from="started", move_to="tested")

    # For updating existing extractor: deactivate replication for old extractor
    if update:
        file_handler.deactivate_replication(src_env=Env.PROD, dst_env=Env.TEST, move_from="started", move_to="tested")

    # User guidance: remind to deploy new / updated extractor to development environment
    workflow_stage = "updated" if update else "new"
    cprint(
        f"Please deploy your {workflow_stage} extractor towards your testing environment by "
        f'populating the data set "{extractor_id}" in "{tenant_info[Env.TEST].name}" with the {workflow_stage} '
        f"extractor. Further you have to deactivate the extractor into the development environment "
        f'"{tenant_info[Env.DEV].name}".',
        "yellow",
    )
    click.confirm(
        f"Once you have done both, hit enter to confirm", show_default=False,
    )

    # Finally, check for rollback of data replications towards dev-tenant
    rollback = check_for_rollback(tenant=test_tenant, stage="tested", update=update, file_handler=file_handler)

    if not rollback:
        s = "an updated" if update else "a new"
        cprint(
            f"Congratulations! You completed all steps to test {s} extractor. Now you can continue with the deploy "
            "step.",
            "green",
        )


def deploy_extractor(tenant_info: Dict[Env, CDFInfo], base_tenant: str, ui: UserInterface, update: bool):
    """
    Guides the user through either
    1) Deploy a new extractor (update: False) or
    2) Deploy an updated, existing extractor (update: True)
    Additionally, the necessary data replications are set up, and redundant data replications are deactivated


    Parameters:
        tenant_info(Dict[str, CDFInfo]): Dictionary with CDF tenant names and clients for each environment)
        base_tenant (str):
        ui (UserInterface):
        update (bool): True for "update extractor" workflow, False for "new extractor" workflow
    """
    test_tenant = tenant_info[Env.TEST].name
    prod_tenant = tenant_info[Env.PROD].name

    # Get extractor name
    read_extractor_location = "prod-test" if update else "prod-dev"
    try:
        extractor_options = get_extractor_names(location=Path(base_tenant, read_extractor_location, "tested"))
        extractor_id = ui.choose_option(prompt="Choose extractor to deploy", options=extractor_options)
    except FileNotFoundError:
        cprint("There is no extractor to deploy. Please select another option.", "red")
        return

    # Check if necessary data sets exist
    if not does_data_set_exist(extractor_id, tenant_info[Env.PROD]):
        create_data_set(extractor_id, tenant_info[Env.PROD])
    while not does_data_set_exist(extractor_id, tenant_info[Env.TEST]):
        cprint(f"There is no extractor {extractor_id} in {test_tenant}.", "red")
        click.confirm(
            "Please test new extractor first.", show_default=False,
        )
        test_new_extractor(tenant_info, base_tenant, ui)

    # Move files from prod-dev/tested -> prod-dev/deactivated and deploy new extractor on prod-test
    schedule = ui.choose_option(
        prompt=f'How often should the data from the extractor "{extractor_id}" be replicated in production?',
        options=SCHEDULE_OPTIONS,
    )
    file_handler = ExtractorWorkflowFileHandler(base_tenant=base_tenant, extractor_id=extractor_id, schedule=schedule)
    file_handler.generate_config_files(
        data_sets_to_replicate=[extractor_id], src_env=Env.PROD, dst_env=Env.TEST, stage="deployed"
    )
    file_handler.activate_replication(src_env=Env.PROD, dst_env=Env.TEST, stage="deployed")
    file_handler.deactivate_replication(src_env=Env.PROD, dst_env=Env.DEV, move_from="tested", move_to="deactivated")

    # For updating existing extractor: move already existing config files
    timestamp = get_utc_timestamp_as_str()
    if update:
        file_handler.move_folder_locally(
            src_env=Env.PROD, dst_env=Env.TEST, move_from="tested", move_to="deactivated", timestamp=timestamp,
        )
        file_handler.move_folder_on_gcp(
            src_env=Env.PROD, dst_env=Env.TEST, move_from="tested", move_to="deactivated", timestamp=timestamp,
        )

    cprint(f'The replication of "{extractor_id}" from "{prod_tenant}" to "{test_tenant}" is activated.', "green")
    print_monitor_replication()
    cprint(
        f'Now you have to develop your extractor towards your production environment by populating "{extractor_id}" in '
        f'"{prod_tenant}" with the new extractor. Further you have to deactivate the extractor into the testing '
        f"environment",
        "yellow",
    )
    click.confirm("Once you have done both, hit enter to confirm", show_default=False)
    s1 = "an updated" if update else "a new"
    s2 = "updated" if update else "new"
    cprint(
        f"Congratulations! You completed all steps to deploy {s1} extractor. Your {s2} extractor is now running in "
        "production.",
        "green",
    )


def get_tenant_names_extractor_workflow(ui) -> Dict:
    """
    Asks the user to choose the naming convention of the tenants for the extractor workflow:
    Option 1): <project>-dev, <project>-test, <project>-prod
    Option 2): <project>-dev, <project>-test, <project>
    and returns the chosen option as a dictionary

    Parameters:
        ui (UserInterface):
    """
    base_tenant = prompt_for_base_info()
    dev_tenant = f"{base_tenant}-{Env.DEV}"
    test_tenant = f"{base_tenant}-{Env.TEST}"
    prod_tenant_0 = f"{base_tenant}-{Env.PROD}"
    prod_tenant_1 = f"{base_tenant}"
    tenant_names_option_0 = f"dev: {dev_tenant}, test: {test_tenant}, prod: {prod_tenant_0}"
    tenant_names_option_1 = f"dev: {dev_tenant}, test: {test_tenant}, prod: {prod_tenant_1}"
    tenant_names_option_custom = "Custom tenant names"
    tenant_names_choice = ui.choose_option(
        prompt="Which naming convention do your CDF tenants follow",
        options=[tenant_names_option_0, tenant_names_option_1, tenant_names_option_custom],
    )
    if tenant_names_choice == tenant_names_option_custom:
        dev_tenant = click.prompt("CDF development tenant name")
        test_tenant = click.prompt("CDF test tenant name")
        prod_tenant = click.prompt("CDF production tenant name")
    elif tenant_names_choice == tenant_names_option_0:
        prod_tenant = prod_tenant_0
    elif tenant_names_choice == tenant_names_option_1:
        prod_tenant = prod_tenant_1
    return {Env.BASE: base_tenant, Env.DEV: dev_tenant, Env.TEST: test_tenant, Env.PROD: prod_tenant}


def prompt_for_base_info() -> str:
    """
    Asks for the base name of the tenant
    """
    tenant = click.prompt("Name of your project / customer")

    return tenant


def get_tenant_names_data_replication() -> Dict:
    """
    Requests source and destination tenant names for the custom data replication workflow and returns both together in
    a dictionary
    """
    src_tenant = click.prompt("Name of the source tenant")
    dst_tenant = click.prompt("Name of the destination tenant")
    tenant_names = {Env.SRC: src_tenant, Env.DST: dst_tenant}
    return tenant_names


def get_tenant_info(
    secret_manager_client: SecretManagerServiceClient, tenant_names: Dict, ui: UserInterface
) -> Optional[Dict[Env, CDFInfo]]:
    """
    Collects api-key and client information about all three CDF tenants and returns them as a Dictionary

    Parameters:
        secret_manager_client (SecretManagerServiceClient):
        tenant_names (Dict): the CDF tenant names for the data replication
        ui (UserInterface):
    """
    cdf_info_dict = {}

    for env, tenant_name in tenant_names.items():
        while True:
            api_key = get_cdf_api_key(client=secret_manager_client, tenant=tenant_name, ui=ui)
            if not api_key:
                return None
            cognite_client = CogniteClient(api_key=api_key, client_name="forge-replicator", project=tenant_name)
            key_valid = is_api_key_valid(client=cognite_client, tenant=tenant_name)
            if key_valid:
                break
            else:
                cprint(
                    f'Invalid api-key for tenant "{tenant_name}". Either there exists no such CDF tenant or the '
                    f"stored api-key is not valid.",
                    "red",
                )
                alt_1 = "Enter tenant names again"
                alt_2 = f'Update api-key for tenant "{tenant_name}" in GCP'
                choice = ui.choose_option("Please choose", options=[alt_1, alt_2])
                if choice == alt_1:
                    return None
                if choice == alt_2:
                    upload_api_key_to_gcp(client=secret_manager_client, tenant=tenant_name, update=True)
        cdf_info_dict[env] = CDFInfo(tenant_name, api_key, cognite_client)

    return cdf_info_dict


def get_invalid_data_sets(data_sets: List[str], tenant_info: CDFInfo) -> List[str]:
    """
    Checks if a list of data sets exist in a given tenant and returns those that don't exist

    Parameters:
        data_sets (List[str]): List of external ids
        tenant_info (CDF_Info): tenant information such as name, api-key and CDF client
    """
    return list(filter(lambda ds: not does_data_set_exist(ds, tenant_info), data_sets))


def does_data_set_exist(data_set: str, tenant_info: CDFInfo) -> bool:
    """
    Checks if a single data set exists in a given tenant

    Parameters:
        data_set: external_id
        tenant_info (CDF_Info): tenant information such as name, api-key and CDF client
    """
    return tenant_info.client.data_sets.retrieve(external_id=data_set) is not None


def create_data_set(data_set: str, tenant_info: CDFInfo) -> None:
    """
    Creates a data set with the same name and external_id on a given tenant

    Parameters:
        data_set (str): Will be used for both name and external_id of the new data_ser
        tenant_info (CDF_Info): tenant information such as name, api-key and CDF client
    """
    tenant_info.client.data_sets.create(DataSet(name=data_set, external_id=data_set))
    logger.info(f"Created data set {data_set} on tenant {tenant_info.name}")


def is_api_key_valid(client: CogniteClient, tenant: str) -> bool:
    """
    Checks if an api_key is valid for a given tenant

    Parameters:
        client (CogniteClient):
        tenant (str): CDF tenant name
    """
    status = client.login.status()
    return status.logged_in and status.project == tenant


def get_cdf_api_key(client: SecretManagerServiceClient, tenant: str, ui: UserInterface) -> Optional[str]:
    """
    Reads the CDF api-key for a given tenant from the GCP Secret Manager. If no key found, asks to abort

    Parameters:
        client (SecretManagerServiceClient):
        tenant (str): CDF tenant name
        ui (UserInterface):
    """
    while True:
        project_id = GCP_PROJECT_NAME
        secret_id = tenant
        name = f"projects/{project_id}/secrets/{secret_id}/versions/{'latest'}"
        try:
            response = client.access_secret_version(request={"name": name})
            payload = response.payload.data.decode()
            gcp_iam_grant_access(client=client, secret_id=tenant)
            return payload
        except NotFound:
            cprint(f'Could not access api-key for tenant "{tenant}" from GCP', "red")
            alt_1 = "Enter tenant names again"
            alt_2 = f'Upload api-key for tenant "{tenant}" as a secret to GCP'
            choice = ui.choose_option("Please choose", options=[alt_1, alt_2])
            if choice == alt_1:
                return None
            if choice == alt_2:
                upload_api_key_to_gcp(tenant=tenant, client=client)


def upload_api_key_to_gcp(tenant: str, client: SecretManagerServiceClient, update=False):
    """
    Requests and uploads CDF api-key as a secret to GCP with the correct access rights for the beam replication engine

    Parameters:
        tenant (str): CDF tenant name
        client (SecretManagerServiceClient):
        update (bool): whether secret is only updated or newly created
    """
    if not update:
        gcp_create_secret(client=client, secret_id=tenant)
        gcp_iam_grant_access(client=client, secret_id=tenant)
    api_key = click.prompt(
        f'Please paste the api-key for the tenant "{tenant}". Make sure that this api-key has the '
        f"necessary permissions (i.e. read / write for the desired data types and data sets)"
    )
    gcp_add_secret_version(client=client, secret_id=tenant, payload=api_key)


def get_extractor_names(location: Path) -> List[str]:
    """
    Reads and returns all folder names on a given Path. This way the available extractors can be read

    Parameters:
         location (Path): path within the project folder
    """
    directory = EW_OUTPUT_DIR / location
    extractor_names = [x.name for x in directory.iterdir() if x.is_dir()]
    if not extractor_names:
        # No extractors in this directory
        raise FileNotFoundError
    return extractor_names


def request_cdf_api_hosts() -> Dict:
    """
    Requests CDF source and destination api hosts and returns them as a dictionary
    """
    if click.confirm("Do you want to enter custom CDF api hosts? (Default: https://api.cognitedata.com)"):
        src_host = request_host("source")
        dst_host = request_host("target")
        return {Env.SRC: src_host, Env.DST: dst_host}
    else:
        return {Env.SRC: DEFAULT_CDF_API_HOST, Env.DST: DEFAULT_CDF_API_HOST}


def request_host(env: str) -> str:
    """
    Requests host and checks that it matches the RegEx

    Parameters:
        env (str): to let the user know for which environment the host is requested
    """
    while True:
        host = click.prompt(f"Please enter the {env} host")
        if re.match(CDF_API_HOST_REGEX, host):
            return host
        else:
            cprint(
                f"The hosting location must be a valid CDF cluster URL (for instance {DEFAULT_CDF_API_HOST})", "yellow"
            )


def check_for_rollback(tenant: str, stage: str, update: bool, file_handler: ExtractorWorkflowFileHandler):
    """
    Asks the user to perform a rollback, in case the new or updated extractor does not work as intended

    Parameters:
        tenant(str): the name of the tenant to check whether rollback desired
        stage (str): from which stage of the workflow the function is called: 'started' or 'tested'
        update (bool): True if called within update_extractor
        file_handler (ExtractorWorkflowFileHandler):
    """
    rollback = False
    extractor_age = "updated" if update else "new"
    while not click.confirm(
        f'Finally, you need to make sure that the {extractor_age} extractor "{file_handler.extractor_id}" works as '
        f'intended. To do so check if the data set "{file_handler.extractor_id}" in "{tenant}" contains the expected '
        f"data. If you are satisfied, hit enter to continue. Otherwise you can deprecate the current "
        f'work on extractor "{file_handler.extractor_id}" by entering "N" ',
        show_default=False,
        default=True,
    ):
        if click.confirm(
            f"Are you sure you want to deprecate the current work on {file_handler.extractor_id} and rollback to the "
            f"previous stage? "
        ):
            rollback = True
            file_handler.deactivate_replication(
                src_env=Env.PROD, dst_env=Env.DEV, move_from=stage, move_to="deprecated"
            )
            if update:
                file_handler.move_folder_locally(
                    src_env=Env.PROD, dst_env=Env.TEST, move_from=stage, move_to="deployed"
                )
                file_handler.move_folder_on_gcp(src_env=Env.PROD, dst_env=Env.TEST, move_from=stage, move_to="deployed")
                if stage == "tested":
                    # Reactivate replication
                    file_handler.activate_replication(src_env=Env.PROD, dst_env=Env.TEST, stage="deployed")
                    pass

            cprint(
                f'Successfully deprecated work on extractor "{file_handler.extractor_id}" and rolled back to previous '
                f"stage",
                "green",
            )
            cprint(f'Make sure to disable the {extractor_age} extractor towards "{tenant}".', "yellow")
            click.confirm("Hit enter to confirm", show_default=False)
            break

    return rollback


def does_job_id_exist(job_id: str, src_tenant: Env, dst_tenant: Env, data_type: str):
    path = CR_OUTPUT_DIR / f"{src_tenant}-{dst_tenant}" / f"{data_type}"
    if path.exists():
        existing_job_ids = [d.name for d in path.iterdir()]
        if job_id in existing_job_ids:
            return True
    return False


def gcp_create_secret(client: SecretManagerServiceClient, secret_id: str):
    """
    Create a new secret with the secret_id as name. A secret is a logical wrapper around a collection of secret
    versions. Secret versions hold the actual secret material.

    Parameters:
        client (SecretManagerServiceClient): GCP client
        secret_id (str): the name of the secret. Should be the tenant name
    """
    client.create_secret(
        request={
            "parent": f"projects/{GCP_PROJECT_NAME}",
            "secret_id": secret_id,
            "secret": {"replication": {"automatic": {}}},
        }
    )


def gcp_add_secret_version(client: SecretManagerServiceClient, secret_id: str, payload):
    """
    Add a new secret version to the given secret with the provided payload, i.e. the api-key

    Parameters:
        client (SecretManagerServiceClient): GCP client
        secret_id (str): the name of the secret. Should be the tenant name
        payload (str): the CDF api-key
    """
    parent = client.secret_path(GCP_PROJECT_NAME, secret_id)
    payload = payload.encode("UTF-8")
    client.add_secret_version(request={"parent": parent, "payload": {"data": payload}})


def gcp_iam_grant_access(client: SecretManagerServiceClient, secret_id: str, members: List[str] = GS_SERVICE_ACCOUNTS):
    """
    Grant correct access rights for the beam replication engine

    Parameters:
        client (SecretManagerServiceClient): GCP client
        secret_id (str): the name of the secret. Should be the tenant name
        members (List[str]): List of service accounts
    """
    name = client.secret_path(GCP_PROJECT_NAME, secret_id)
    policy = client.get_iam_policy(request={"resource": name})
    policy.bindings.add(role="roles/secretmanager.secretAccessor", members=members)
    client.set_iam_policy(request={"resource": name, "policy": policy})


def handle_default_credentials_error(e: DefaultCredentialsError) -> NoReturn:
    """
    Handles case when GCP Authentication fails due do DefaultCredentialsError. Prints error information and exits.

    Parameters:
        e (DefaultCredentialsError)
    """
    cprint("Couldn't read GCP credentials", "red")
    cprint(f'To learn how to set up authentication, please check the "Configuration" section under {DOC_URL}', "red")
    print_goodbye()
    sys.exit(0)



# import os
# import shutil
# import tempfile
# from os import listdir
# from os.path import isfile, join
# from typing import Dict, List
#
# from azure.devops.connection import Connection
# from azure.devops.exceptions import AzureDevOpsServiceError
# from azure.devops.v5_1.build.models import (
#     AgentPoolQueue,
#     BuildDefinition,
#     BuildDefinitionVariable,
#     BuildRepository,
#     TeamProjectReference,
# )
# from azure.devops.v5_1.git.models import GitRepositoryCreateOptions
# from git import Repo
# from msrest.authentication import BasicAuthentication
#
# import forge_template.handler.git_utils as git_utils
# import forge_template.util.log_utils as print_utils
# from forge_template.handler.handler import BaseHandler
# from forge_template.util.yaml_util import DATABRICKS_CONFIG_NAME, PIPELINES_PATH, POWERBI_CONFIG_NAME
#
# PIPELINE_SERVICES = ["databricks", "powerbi"]
#
#
# class AzureHandler(BaseHandler):
#     """
#     Handler for interacting with Azure DevOps.
#     Currently supporting:
#         - Create Azure Repo
#         - Create Azure Build Definition
#     """
#
#     def __init__(self, name: str, config: Dict):
#         self._databricks_config = config[DATABRICKS_CONFIG_NAME]
#         self._powerbi_config = config[POWERBI_CONFIG_NAME]
#
#         config = config[name]
#         self._organization = config["principal"]
#         self._project_name = config["project_name"]
#         self._repo_name = config["repo_name"]
#
#         url = f"https://dev.azure.com/{self._organization}"
#         credentials = BasicAuthentication("", config["token"])
#         connection = Connection(base_url=url, creds=credentials)
#         self._core_client = connection.clients.get_core_client()
#         self._build_client = connection.clients.get_build_client()
#         self._git_client = connection.clients.get_git_client()
#
#         self._git_tmp_work_dir = tempfile.TemporaryDirectory().name
#         self._local_repo = Repo.init(path=self._git_tmp_work_dir)
#
#         self._project_id: str = ""
#         self._repo_url: str = ""
#         self._repo_ssh_url: str = ""
#         self._definition_ids: Dict[str, str] = {}
#
#         self._build_definitions_to_create: List[_BuildDefinition] = []
#         self._created_build_definitions: List[Dict[str, str]] = []
#
#         self._repos_to_create: List[str] = []
#         self._remotes_to_rename: List[str] = []
#         self._remotes_to_add: List[str] = []
#
#         self._created_repos: Dict[str, str] = {}
#         self._renamed_remotes: List[str] = []
#         self._added_remotes: List[str] = []
#
#         super().__init__(name=name, config=config)
#
#     @property
#     def __project_id(self) -> str:
#         return self._project_id if self._project_id else self.__get_project_id()
#
#     def __get_project_id(self) -> str:
#         project = self._core_client.get_project(project_id=self._project_name)
#         self._project_id = project.id
#         return project.id
#
#     @property
#     def __repo_url(self) -> str:
#         return self._repo_url if self._repo_url else self.__get_repository_urls()
#
#     @property
#     def __repo_ssh_url(self) -> str:
#         return self._repo_ssh_url if self._repo_ssh_url else self.__get_repository_urls(return_ssh=True)
#
#     @property
#     def __repo_id(self) -> str:
#         return self.__get_repository_id()
#
#     @property
#     def __definition_ids(self) -> Dict[str, str]:
#         return self._definition_ids if self._definition_ids else self.__get_existing_definitions(self.__project_id)
#
#     def __get_repository_urls(self, return_ssh: bool = False) -> str:
#         try:
#             repository = self._git_client.get_repository(repository_id=self._repo_name, project=self.__project_id)
#             self._repo_ssh_url = repository.ssh_url
#             self._repo_url = repository.url
#             return repository.ssh_url if return_ssh else repository.url
#         except AzureDevOpsServiceError as e:
#             if e.type_key == "GitRepositoryNotFoundException":
#                 return ""
#             raise e
#
#     def __get_repository_id(self):
#         repository = self._git_client.get_repository(repository_id=self._repo_name, project=self.__project_id)
#         return repository.id
#
#     def create_preview(self) -> None:
#         if not self.__repo_ssh_url:
#             self._repos_to_create.append(self._repo_name)
#         git_utils.check_remotes_to_add
#         (self._local_repo, self._repo_name, self._remotes_to_rename, self._remotes_to_add)
#         self.create_definition_preview(self.__project_id)
#
#     def print_preview(self) -> None:
#         print_utils.print_resource_to_add([self._repo_name], "Repository")
#         print_utils.print_remotes_to_rename(self._remotes_to_rename)
#         print_utils.print_resource_to_add(self._remotes_to_add, "Remote")
#         print_utils.print_resource_to_add(self._build_definitions_to_create, "Build definition")
#
#     def setup(self) -> None:
#         for repo_name in self._repos_to_create:
#             self.create_repo(repo_name, self.__project_id)
#
#         try:
#             remote_name = self._local_repo.remote()
#         except ValueError:
#             remote_name = None
#         if self._repo_ssh_url is not None and remote_name is None:
#             self._local_repo.create_remote("origin", self._repo_ssh_url)
#
#         self.__copy_files_to_repo_dir()
#
#         git_utils.push_content_to_repo(self._local_repo, self._repo_name)
#
#         for definition in self._build_definitions_to_create:
#             self.create_definition(
#                 name=definition.name, variables=definition.build_variables, file_path=definition.file_path
#             )
#
#     def rollback(self) -> None:
#         for definition in self._created_build_definitions:
#             self.delete_definition(definition=definition)
#
#         for repo_id, repo_name in self._created_repos.items():
#             self.delete_repo(name=repo_name, repo_id=repo_id)
#
#         git_utils.remove_remotes(self._added_remotes, self._local_repo)
#         git_utils.rename_remotes(self._renamed_remotes, self._local_repo, [])
#
#     def delete_all_resources(self):
#         self.delete_repo(self._repo_name, self.__repo_id)
#         for def_name, def_id in self.__definition_ids.items():
#             self.delete_definition({"name": def_name, "id": def_id})
#
#     def create_repo(self, repo_name: str, project_id: str) -> None:
#         """
#         Creates a new Azure DevOps Repository
#
#         Args:
#             repo_name: Name of repository
#             project_id: Id of Azure DevOps project where the repository will be created
#         """
#         create_repo = GitRepositoryCreateOptions(name=repo_name)
#         response = self._git_client.create_repository(create_repo, project=project_id)
#         self._created_repos[repo_name] = response.id
#         self._repo_ssh_url = response.ssh_url
#         print_utils.print_success_created(repo_name, "Repository")
#
#     def delete_repo(self, name: str, repo_id: str) -> None:
#         """
#         Deletes an Azure DevOps Repository
#
#         Args:
#             name: Name of repository
#             repo_id: Id of Azure DevOps Repository that will be deleted
#         """
#         self._git_client.delete_repository(repo_id, project=self._project_id)
#         print_utils.print_success_deleted(name, "Repository")
#
#     def create_definition_preview(self, project_id: str):
#         """
#         Checks which build definitions that does not exist. All definition files in the pipelines folder are checked.
#         If a build definition does not exist, it is added to the list of definitions (_build_definitions_to_create)
#         that are to be created.
#         """
#         existing_definitions = self.__get_existing_definitions(project_id=project_id)
#         for file_path in os.listdir(PIPELINES_PATH):
#             service = os.path.splitext(file_path)[0]
#             definition_name = f"{self._project_name}-{service}"
#             if definition_name not in existing_definitions:
#                 self._build_definitions_to_create.append(
#                     _BuildDefinition(
#                         definition_name, self.__get_build_variables(service), os.path.join(PIPELINES_PATH, file_path),
#                     )
#                 )
#
#     def create_definition(self, name: str, variables: Dict[str, BuildDefinitionVariable], file_path: str,) -> None:
#         """
#         Creates a Azure DevOps Build definition
#
#         Args:
#             name: Name of definition
#             variables: Dictionary mapping variable name to its definition (BuildDefinitionVariable)
#             file_path: Path of .yaml/.yml file containing the definition
#         """
#
#         build_repo = BuildRepository(
#             default_branch="refs/heads/master",
#             checkout_submodules=False,
#             name=self._repo_name,
#             url=self.__repo_url,
#             type="TfsGit",
#             properties={"reportBuildStatus": True},
#         )
#
#         queue = AgentPoolQueue(name="Azure Pipelines")
#         triggers = {"settingsSourceType": 2, "batchChanges": False, "triggerType": "continuousIntegration"}
#         project = TeamProjectReference(id=self.__project_id)
#         process = {"type": 2, "yamlFilename": file_path}
#
#         definition = BuildDefinition(
#             name=name,
#             project=project,
#             repository=build_repo,
#             process=process,
#             type="build",
#             variables=variables,
#             queue=queue,
#             triggers=[triggers],
#         )
#
#         response = self._build_client.create_definition(definition, self.__project_id)
#         self._created_build_definitions.append({"name": name, "id": response.id})
#         print_utils.print_success_created(name, "Build definition")
#
#     def delete_definition(self, definition: Dict[str, str]) -> None:
#         """
#         Deletes a definition
#
#         Args:
#             definition: Definition to be deleted as a dictionary with keys:
#                 ``"id"``
#                 Id of definition to delete (str)
#
#                 ``"name"``
#                 Name of definition to delete (str)
#         """
#         self._build_client.delete_definition(self.__project_id, definition["id"])
#         print_utils.print_success_deleted(definition["name"], "Build definition")
#
#     def __get_build_variables(self, service_name: str) -> Dict[str, BuildDefinitionVariable]:
#         if service_name == "databricks":
#             return {
#                 "DatabricksHost": BuildDefinitionVariable(value=self._databricks_config["host"]),
#                 "DatabricksToken": BuildDefinitionVariable(value=self._databricks_config["token"], is_secret=True),
#                 "DatabricksNotebookFolder": BuildDefinitionVariable(value="databricks"),
#                 "DatabricksWorkspaceFolder": BuildDefinitionVariable(
#                     value=self._databricks_config["production"]["workspace_path"]
#                 ),
#             }
#         elif service_name == "powerbi":
#             return {
#                 "POWERBI_TENANT_ID": BuildDefinitionVariable(value=self._powerbi_config["tenant_id"]),
#                 "POWERBI_APPLICATION_ID": BuildDefinitionVariable(value=self._powerbi_config["application_id"]),
#                 "POWERBI_TOKEN": BuildDefinitionVariable(value=self._powerbi_config["token"], is_secret=True),
#                 "POWERBI_WORKSPACE": BuildDefinitionVariable(value=self._powerbi_config["production"]["name"]),
#             }
#         else:
#             raise RuntimeError(f"Unexpected
#             service name. Got {service_name}, but expected on of [databricks, powerbi]")
#
#     def __get_existing_definitions(self, project_id: str) -> Dict[str, str]:
#         """
#         Fetches the existing build definitions for the current project id
#
#         Returns: A list of definition names
#         """
#         definitions = self._build_client.get_definitions(project_id).value
#         return {definition.name: definition.id for definition in definitions}
#
#     def __copy_files_to_repo_dir(self) -> None:
#         working_dir = os.getcwd()
#
#         pipelines_path = os.path.join(working_dir, "pipelines")
#         template_databricks_path = os.path.join(working_dir, "template", "databricks")
#         template_powerbi_path = os.path.join(working_dir, "template", "powerbi")
#         template_infra_path = os.path.join(working_dir, "template", "infra_scripts")
#         config_path = os.path.join(working_dir, "config.yaml")
#         requirements_path = os.path.join(working_dir, "requirements.txt")
#
#         self.__handle_copy_dir_files(pipelines_path, "pipelines")
#         self.__handle_copy_dir_files(template_databricks_path, "databricks")
#         self.__handle_copy_dir_files(template_powerbi_path, "powerbi")
#         self.__handle_copy_dir_files(template_infra_path, "infra_scripts")
#
#         shutil.copy(config_path, self._git_tmp_work_dir)
#         shutil.copy(requirements_path, self._git_tmp_work_dir)
#
#     def __handle_copy_dir_files(self, source_dir: str, file_type: str) -> None:
#         files = [
#             join(source_dir, f) for f in listdir(source_dir) if isfile(join(source_dir, f)) and f != "." and f != ".."
#         ]
#         dest_path = os.path.join(self._git_tmp_work_dir, file_type)
#         os.mkdir(dest_path)
#         for file in files:
#             shutil.copy(file, dest_path)
#
#
# class _BuildDefinition:
#     def __init__(self, name: str, build_variables: Dict[str, BuildDefinitionVariable], file_path: str):
#         self.name = name
#         self.build_variables = build_variables
#         self.file_path = file_path



import time

from forge_template.sa_constants import GCP_PROJECT_NAME


def get_utc_timestamp_as_str(f="%y%m%d%H%M%S") -> str:
    return time.strftime(f, time.gmtime())


def print_monitor_replication():
    print(
        f"To monitor the replication job, visit: https://console.cloud.google.com/dataflow/jobs?project="
        f"{GCP_PROJECT_NAME} \n (It may take some time until the job appears)"
    )



GCP_PROJECT_NAME = "cognite-sa-sandbox"
COMPOSER_BUCKET = "europe-west1-composer-ccb9555e-bucket"
JOB_DEFINITIONS_BUCKET = "sa-sandbox-training"
JOB_CONFIGS_BUCKET = "sa-forge-replicator"
SCHEDULE_OPTIONS = ["@once", "@hourly", "@daily", "@weekly", "@monthly"]
DATA_TYPE_OPTIONS = ["assets", "ts", "events"]
GS_SERVICE_ACCOUNTS = ["serviceAccount:719303257135-compute@developer.gserviceaccount.com"]
DEFAULT_CDF_API_HOST = "https://api.cognitedata.com"
CDF_API_HOST_REGEX = r"https:\/\/[a-zA-Z0-9][a-zA-Z0-9-]+[.]cognitedata[.]com$"
DOC_URL = "https://cognitedata.atlassian.net/wiki/spaces/FORGE/pages/1720156342/Extractor+workflow+data" "+replication "



import logging
import time
from enum import Enum
from pathlib import Path
from typing import Dict, List

from google.cloud import storage

from forge_template.config_generator import ConfigGenerator, FileType
from forge_template.paths import (
    CR_GCP_CONFIG_PATH,
    CR_GCP_SCHEDULE_PATH,
    CR_OUTPUT_DIR,
    CR_SCHEDULE_TEMPLATE,
    EW_GCP_CONFIG_PATH,
    EW_GCP_SCHEDULE_PATH,
    EW_OUTPUT_DIR,
    EW_SCHEDULE_TEMPLATE,
    JOB_CONFIG_SCHEMA_PATH,
    JOB_CONFIG_TEMPLATE,
)
from forge_template.sa_constants import (
    COMPOSER_BUCKET,
    DATA_TYPE_OPTIONS,
    GCP_PROJECT_NAME,
    JOB_CONFIGS_BUCKET,
    JOB_DEFINITIONS_BUCKET,
)
from forge_template.util.file_util import load_yaml, move_directory, save_toml
from forge_template.util.jinja2_util import render
from forge_template.util.sa_util import get_utc_timestamp_as_str

logger = logging.getLogger(__file__)


class Env(Enum):
    BASE = "base"
    DEV = "dev"
    TEST = "test"
    PROD = "prod"
    SRC = "src"
    DST = "dst"

    def __str__(self):
        return self.value


class FileHandler:
    def __init__(self) -> None:
        self.gcp_client = storage.Client()

    def upload_blob(self, bucket_name: str, source_file_name: str, destination_blob_name: Path) -> None:
        """
        Uploads a file to the gcp bucket.

        Parameters:
            bucket_name (str): GCP bucket name
            source_file_name (str): local path to file
            destination_blob_name (Path): destination path
        """
        bucket = self.gcp_client.bucket(bucket_name)
        blob = bucket.blob(str(destination_blob_name))

        blob.upload_from_filename(source_file_name)

    def delete_blob(self, bucket_name: str, blob_name: str) -> None:
        """
        Deletes a folder from the gcp bucket.

        Parameters:
            bucket_name (str): GCP bucket name
            blob_name (str): path of object to be deleted
        """
        bucket = self.gcp_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob.delete()

    def copy_blob(
        self, bucket_name: str, blob_name: str, destination_bucket_name: str, destination_blob_name: Path
    ) -> None:
        """
        Copies a blob from one bucket to another with a new name.

        Parameters:
            bucket_name (str): GCP bucket name (source)
            blob_name (str): object path (source)
            destination_bucket_name (str): GCP bucket name (destination)
            destination_blob_name (Path): object path (target)
        """
        source_bucket = self.gcp_client.bucket(bucket_name)
        source_blob = source_bucket.blob(blob_name)
        destination_bucket = self.gcp_client.bucket(destination_bucket_name)

        source_bucket.copy_blob(source_blob, destination_bucket, str(destination_blob_name))

    def list_blobs_with_prefix(self, bucket_name: str, prefix: Path) -> List[str]:
        """
        Lists all the blobs in the bucket that begin with the prefix.

        Parameters:
            bucket_name (str): GCP bucket name (source)
            prefix (Path): prefix of path to object
        """
        return [blob.name for blob in self.gcp_client.list_blobs(bucket_name, prefix=str(prefix))]


class ExtractorWorkflowFileHandler(FileHandler):
    def __init__(self, base_tenant: str, extractor_id: str, schedule: str = "") -> None:
        super().__init__()
        self.base_tenant = base_tenant
        self.extractor_id = extractor_id
        self.schedule = schedule

        self.config_files_path = Path()

    def generate_config_files(self, data_sets_to_replicate: List[str], src_env: Env, dst_env: Env, stage: str) -> None:
        """
        Generates all 6 config files: jobs and schedules for assets, time series and events.

        Parameters:
            data_sets_to_replicate (List[str]): list of external ids
            src_env (Env): The environment of the data replication source.
            dst_env (Env): The environment of the data replication destination.
            stage (str): in which stage the config files are generated.
        """
        dst_path = EW_OUTPUT_DIR / self.base_tenant / f"{src_env}-{dst_env}" / stage / self.extractor_id
        dst_path.mkdir(parents=True, exist_ok=True)
        self.config_files_path = dst_path

        self.generate_job_config_files(dst_path, data_sets_to_replicate)
        self.generate_schedule_config_files(path=dst_path, src_env=src_env, dst_env=dst_env)

    @staticmethod
    def generate_job_config_files(path: Path, data_sets_to_replicate: List[str]) -> None:
        """
        Generates asset, time series and events config file and stores them under path

        Parameters:
            path (Path): destination path of the config files
            data_sets_to_replicate (List[str]): list of external ids
        """
        config_dict = load_yaml(JOB_CONFIG_TEMPLATE)

        for job in DATA_TYPE_OPTIONS:
            job_config_dict = config_dict[job]
            job_config_dict["allowList"]["dataSetExternalId"] = data_sets_to_replicate
            full_path = path / f"{job}-config.toml"
            save_toml(full_path, job_config_dict)

    def generate_schedule_config_files(self, path: Path, src_env: Env, dst_env: Env) -> None:
        """
        Generation of schedule files

        Parameters:
            path (Path): destination path of schedule files
            src_env (str): The environment of the data replication source.
            dst_env (str): The environment of the data replication destination.
        """
        for job in DATA_TYPE_OPTIONS:
            content = {
                "job": job,
                "src_env": src_env.value,
                "dst_env": dst_env.value,
                "gcp_project_name": GCP_PROJECT_NAME,
                "job_definitions_bucket": JOB_DEFINITIONS_BUCKET,
                "job_configs_bucket": JOB_CONFIGS_BUCKET,
                "tenant": self.base_tenant,
                "extractor_id": self.extractor_id,
                "schedule": self.schedule,
                "dag_name": f"replicate_{job}_on_{self.base_tenant}_{src_env}_{dst_env}_{self.extractor_id}_"
                f"{time.strftime('%y%m%d%H%M%S', time.gmtime())}",
            }
            output_path = path / f"{job}-schedule.py"
            render(content, EW_SCHEDULE_TEMPLATE, output_path)

    def activate_replication(self, src_env: Env, dst_env: Env, stage: str) -> None:
        """
        Activates the replication by uploading the corresponding config files to gcp

        Parameters:
            src_env (Env): The environment of the data replication source.
            dst_env (Env): The environment of the data replication destination.
            stage (str): in which stage the config files are generated / the folder name of the config files
        """
        # avoid temporary and hidden files
        config_files = [f for f in self.config_files_path.iterdir() if f.is_file() and not f.name.startswith(".")]

        path_suffix = Path(self.base_tenant) / f"{src_env}-{dst_env}" / stage / self.extractor_id
        gcp_config_path = EW_GCP_CONFIG_PATH / path_suffix
        gcp_schedule_path = EW_GCP_SCHEDULE_PATH / path_suffix
        for f in config_files:
            # Upload job configs and schedules
            self.upload_blob(
                bucket_name=JOB_CONFIGS_BUCKET, source_file_name=str(f), destination_blob_name=gcp_config_path / f.name,
            )
            # Upload only schedules to COMPOSER_BUCKET
            if "schedule" in f.name:
                self.upload_blob(
                    bucket_name=COMPOSER_BUCKET,
                    source_file_name=str(f),
                    destination_blob_name=gcp_schedule_path / f.name,
                )

    def deactivate_replication(self, src_env: Env, dst_env: Env, move_from: str, move_to: str) -> None:
        """
        Deactivates the replication by deleting the schedule files from COMPOSER_BUCKET and updating the folder
        structure locally and on JOB_CONFIGS_BUCKET

        Parameters:
            src_env (Env): The environment of the data replication source.
            dst_env (Env): The environment of the data replication destination.
            move_from (str): Current location of the config files of the replication to be deactivated
            move_to (str): New location after replication has been deactivated.
        """
        timestamp = ""
        if move_to in ["deactivated", "deprecated"]:
            timestamp = get_utc_timestamp_as_str()
        self.delete_schedule_files(src_env=src_env, dst_env=dst_env, stage=move_from)
        self.move_folder_locally(
            src_env=src_env, dst_env=dst_env, move_from=move_from, move_to=move_to, timestamp=timestamp
        )
        self.move_folder_on_gcp(
            src_env=src_env, dst_env=dst_env, move_from=move_from, move_to=move_to, timestamp=timestamp
        )

    def delete_schedule_files(self, src_env: Env, dst_env: Env, stage: str) -> None:
        """
        Deletes schedule files in COMPOSER_BUCKET/dags

        Parameters:
            src_env (str): The environment of the data replication source.
            dst_env (str): The environment of the data replication destination.
            stage (str): The stage in which the files are before deletion, i.e. "started" or "tested"
        """
        path = EW_GCP_SCHEDULE_PATH / self.base_tenant / f"{src_env}-{dst_env}" / stage / self.extractor_id
        files_to_delete = self.list_blobs_with_prefix(COMPOSER_BUCKET, prefix=path)
        for f in files_to_delete:
            self.delete_blob(bucket_name=COMPOSER_BUCKET, blob_name=f)

    def move_folder_locally(
        self, src_env: Env, dst_env: Env, move_from: str, move_to: str, timestamp: str = ""
    ) -> None:
        """
        Moves a folder locally

        Parameters:
            src_env (Env): The environment of the data replication source.
            dst_env (Env): The environment of the data replication destination.
            move_from (str): Current location of the config files of the replication to be moved
            move_to (str): New location after replication has been moved.
            timestamp (str): timestamp to be added to folder name. If empty, no timestamp is added
        """
        base_path = Path(EW_OUTPUT_DIR) / self.base_tenant / f"{src_env}-{dst_env}"
        src_path = base_path / move_from / self.extractor_id if move_from else base_path / self.extractor_id
        folder_name = f"{self.extractor_id}_{timestamp}" if timestamp else self.extractor_id
        dst_path = base_path / move_to / folder_name
        move_directory(src=src_path, dst=dst_path)

        self.config_files_path = dst_path

    def move_folder_on_gcp(self, src_env: Env, dst_env: Env, move_from: str, move_to: str, timestamp: str = "") -> None:
        """
        Moves a folder in the GCP project to deactivated or started

        Parameters:
            src_env (Env): The environment of the data replication source.
            dst_env (Env): The environment of the data replication destination.
            move_from (str): Current location of the config files of the replication to be moved
            move_to (str): New location after replication has been moved.
            timestamp (str): timestamp to be added to folder name. If empty, no timestamp is added
        """
        folder_name = f"{self.extractor_id}_{timestamp}" if timestamp else self.extractor_id

        # 1) Move on config bucket
        base_config_path = EW_GCP_CONFIG_PATH / self.base_tenant / f"{src_env}-{dst_env}"
        old_config_path = (
            base_config_path / move_from / self.extractor_id if move_from else base_config_path / self.extractor_id
        )
        new_config_path = base_config_path / move_to / folder_name
        config_files_to_move = self.list_blobs_with_prefix(JOB_CONFIGS_BUCKET, prefix=old_config_path)
        for f in config_files_to_move:
            self.copy_blob(
                bucket_name=JOB_CONFIGS_BUCKET,
                blob_name=f,
                destination_bucket_name=JOB_CONFIGS_BUCKET,
                destination_blob_name=new_config_path / Path(f).name,
            )
            self.delete_blob(bucket_name=JOB_CONFIGS_BUCKET, blob_name=f)

        # 2) Move on schedule bucket
        base_schedule_path = EW_GCP_SCHEDULE_PATH / self.base_tenant / f"{src_env}-{dst_env}"
        old_schedule_path = (
            base_schedule_path / move_from / self.extractor_id if move_from else base_schedule_path / self.extractor_id
        )
        new_schedule_path = base_schedule_path / move_to / folder_name
        schedule_files_to_move = self.list_blobs_with_prefix(COMPOSER_BUCKET, prefix=old_schedule_path)
        for f in schedule_files_to_move:
            self.copy_blob(
                bucket_name=COMPOSER_BUCKET,
                blob_name=f,
                destination_bucket_name=COMPOSER_BUCKET,
                destination_blob_name=new_schedule_path / Path(f).name,
            )
            self.delete_blob(bucket_name=COMPOSER_BUCKET, blob_name=f)


class CustomReplicatorFileHandler(FileHandler):
    def __init__(self, data_type: str, schedule: str, tenant_names: Dict, cdf_api_hosts: Dict, job_id: str) -> None:
        super().__init__()
        self.data_type = data_type
        self.schedule = schedule
        self.src_tenant = tenant_names[Env.SRC]
        self.dst_tenant = tenant_names[Env.DST]
        self.src_host = cdf_api_hosts[Env.SRC]
        self.dst_host = cdf_api_hosts[Env.DST]

        self.base_path = Path(f"{self.src_tenant}-{self.dst_tenant}", f"{data_type}", f"{job_id}")

        self.local_base_path = CR_OUTPUT_DIR / self.base_path
        self.local_job_config_path = self.local_base_path / f"{job_id}.toml"
        self.local_schedule_path = self.local_base_path / f"{job_id}-schedule.py"
        self.local_base_path.mkdir(exist_ok=True, parents=True)

        self.config_generator = ConfigGenerator(FileType.TOML)

    def generate_job_config(self) -> None:
        """
        Generates, validates and saves the config file for the data replication based on the user's input
        """
        self.config_generator.set_schema_info(
            output_path=self.local_job_config_path,
            schema_path=JOB_CONFIG_SCHEMA_PATH,
            key=self.data_type,
            post_transforms=[],
        )
        self.config_generator.generate()

        self.fix_config_keys()

    def generate_schedule(self) -> None:
        """
        Generates and saves the schedule file for the data replication based on the user specified schedule
        """
        content = {
            "data_type": self.data_type,
            "src_tenant": self.src_tenant,
            "dst_tenant": self.dst_tenant,
            "src_host": self.src_host,
            "dst_host": self.dst_host,
            "schedule": self.schedule,
            "path": self.base_path,
            "job_config_name": self.local_job_config_path.name,
            "dag_name": f"replicate_{self.data_type}_from_{self.src_tenant}_to_{self.dst_tenant}_"
            f"{time.strftime('%y%m%d%H%M%S', time.gmtime())}",
            "gcp_project_name": GCP_PROJECT_NAME,
            "job_definitions_bucket": JOB_DEFINITIONS_BUCKET,
            "job_configs_bucket": JOB_CONFIGS_BUCKET,
        }
        render(content, CR_SCHEDULE_TEMPLATE, self.local_schedule_path)

    def upload_configs_to_gcp(self) -> None:
        """
        Activates the replication by uploading config and schedule files to GCP
        """
        self.upload_blob(
            bucket_name=JOB_CONFIGS_BUCKET,
            source_file_name=str(self.local_job_config_path),
            destination_blob_name=CR_GCP_CONFIG_PATH / self.base_path / self.local_job_config_path.name,
        )
        self.upload_blob(
            bucket_name=JOB_CONFIGS_BUCKET,
            source_file_name=str(self.local_schedule_path),
            destination_blob_name=CR_GCP_CONFIG_PATH / self.base_path / self.local_schedule_path.name,
        )
        self.upload_blob(
            bucket_name=COMPOSER_BUCKET,
            source_file_name=str(self.local_schedule_path),
            destination_blob_name=CR_GCP_SCHEDULE_PATH / self.base_path / self.local_schedule_path.name,
        )

    def fix_config_keys(self) -> None:
        """
        The config generator generates config files with keys like "[assets.config]". The beam replicator can only
        handle keys of the form "[config]". This function fixes the config files to be processable by
        the beam-replicator.
        """
        self.config_generator.config = self.config_generator.config[self.data_type]



import logging
from typing import Callable, Dict

import click
from ruamel.yaml.parser import ParserError
from ruamel.yaml.scanner import ScannerError

from forge_template.config_generator import ConfigGenerator
from forge_template.constants import Tools
from forge_template.exception.exceptions import AccessDeniedException, ValidationException
from forge_template.exception.handler import (
    handle_access_denied_error,
    handle_file_not_found,
    handle_parser_error,
    handle_validation_error,
)
from forge_template.handler.handler import BaseHandler
from forge_template.tool_info import ToolInfo
from forge_template.user_interface import UserInterface
from forge_template.util.dict_util import assert_same_keys, assert_scope_matches, merge_dictionaries
from forge_template.util.file_util import load_and_validate_yaml


def _instantiate_config(tool_info: ToolInfo) -> Dict:
    configs = {}
    for info in tool_info.schema_info:
        configs[info.output_path] = load_and_validate_yaml(info.output_path, info.schema_path)

    if tool_info.assert_match:
        assert_same_keys(configs)

    if tool_info.assert_scope_match:
        assert_scope_matches(configs)

    merged = merge_dictionaries(list(configs.values()))

    return merged


def _instantiate_handler(tool_info: ToolInfo) -> BaseHandler:
    config = _instantiate_config(tool_info)
    return tool_info.handler(config, tool_info)


def generate(tool_info: ToolInfo) -> str:
    def action():
        is_custom = click.confirm(
            f"Do you want to do a custom configuration of {tool_info.name} (only recommended in very special cases)?",
            default=False,
            show_default=True,
        )

        generator = ConfigGenerator(file_type=tool_info.file_type)
        for schema_info in tool_info.schema_info:
            if tool_info.generate_all or click.confirm(f"Do you want to generate {schema_info.output_path}?"):
                generator.set_schema_info(
                    output_path=schema_info.output_path,
                    schema_path=schema_info.schema_path,
                    key=tool_info.name,
                    post_transforms=schema_info.post_transforms,
                )
                UserInterface.print_ok(f"\nGenerating {schema_info.output_path}")
                generator.generate(is_custom=is_custom)
                generator.preview()
                logging.info("\n")
                if click.confirm("Do you want to save the preview?"):
                    generator.save()
                else:
                    return

    do_action(action)
    return UserInterface.RETURN


def deploy(tool_info: ToolInfo) -> str:
    def action():
        handler = _instantiate_handler(tool_info)
        handler.run_create_preview()
        handler.run_print_preview()
        if click.confirm("Do you want to proceed?", default=False):
            handler.run_setup()
            UserInterface.print_ok(f"Successfully deployed {tool_info.name}")

    do_action(action)
    return UserInterface.RETURN


def delete(tool_info: ToolInfo) -> str:
    def action():
        handler = _instantiate_handler(tool_info)
        handler.run_delete_all_resources()
        UserInterface.print_ok(f"Successfully deleted {tool_info.name}")

    do_action(action)
    return UserInterface.RETURN


def upload_config() -> None:
    def action():
        handler = Tools.DATABRICKS.handler(_instantiate_config(Tools.DATABRICKS), Tools.DATABRICKS)
        handler.upload_config()

    do_action(action)


def do_action(action: Callable[[], None]) -> None:
    try:
        action()
    except FileNotFoundError as e:
        handle_file_not_found(e)
    except (ParserError, ScannerError) as e:
        handle_parser_error(e)
    except ValidationException as e:
        handle_validation_error(e)
    except AccessDeniedException as e:
        handle_access_denied_error(e)



import base64
import os
import sys
from collections import deque
from typing import Dict, List, Tuple
from uuid import uuid1

from databricks_api import DatabricksAPI
from requests.exceptions import HTTPError
from ruamel.yaml import YAML

CONFIG_FILE_PATH = "config.yaml"


def copy_workspace_in_databricks(src: str, dest: str, handler: DatabricksAPI, overwrite: bool = False) -> None:
    print(f"Trying to copy from {src} to {dest} ...")
    if check_path_exists(src, handler):
        exported = handler.workspace.export_workspace(src, format="DBC")
        if overwrite:
            delete_workspace(dest, handler)
        handler.workspace.import_workspace(dest, format="DBC", content=exported["content"])
        print(f"Done copying.")
    else:
        print(f"Workspace with path {src} does not exist.")


def restore_production_workspace(copy_loc: str, dest: str, handler: DatabricksAPI) -> None:
    print("Trying to restore production workspace from copy ...")
    if check_path_exists(copy_loc, handler):
        copy_workspace_in_databricks(copy_loc, dest, handler, overwrite=True)
        print(f"Workspace with path {dest} successfully restored.")
    else:
        print(f"Workspace with path {copy_loc} does not exist. Cannot perform restore. Please do it manually.")


def delete_workspace(path: str, handler: DatabricksAPI) -> None:
    print(f"Trying to delete workspace with path {path} ... ")
    if check_path_exists(path, handler):
        handler.workspace.delete(path, recursive=True)
        print(f"Workspace with path {path} successfully deleted.")
    else:
        print(f"Workspace with path {path} does not exist.")


def check_path_exists(path: str, handler: DatabricksAPI) -> bool:
    try:
        handler.workspace.get_status(path)
        return True
    except HTTPError as e:
        if e.response.status_code == 404:
            return False
        else:
            raise e


def read_config() -> Dict:
    with open(CONFIG_FILE_PATH) as f:
        data = YAML(typ="safe").load(f)["databricks"]

    return data


def get_upload_paths_and_dirs(workspace_path: str) -> Tuple[Tuple[List[str], List[str]], List[str]]:
    stack = deque(["databricks"])
    local_paths = []
    while stack:
        curr_path = stack.pop()
        if os.path.isfile(curr_path):
            local_paths.append(curr_path)
        else:
            sub_paths = list(map(lambda p: os.path.join(curr_path, p), os.listdir(curr_path)))
            stack.extend(sub_paths)

    upload_paths: List[str] = list(map(lambda p: p.replace("databricks", workspace_path), local_paths))
    upload_dirs: List[str] = list(map(lambda p: os.path.dirname(p).replace("databricks", workspace_path), local_paths))
    paths = (local_paths, upload_paths)
    return paths, upload_dirs


def upload_notebooks_and_folders(paths: Tuple[List[str], List[str]], upload_dirs, handler: DatabricksAPI) -> None:
    print("Creating directories ...")
    for d in upload_dirs:
        handler.workspace.mkdirs(d)
        print(f"Created {d} ...")

    print("\nUploading notebooks ...")
    for local_path, upload_path in zip(paths[0], paths[1]):
        with open(local_path, "rb") as f:
            content = f.read()
            encoded_content = base64.b64encode(content).decode()
            handler.workspace.import_workspace(upload_path, language="PYTHON", format="SOURCE", content=encoded_content)
            print(f"Uploaded {local_path} as {upload_path}")


def upload_from_repo(prod_workspace_path: str, handler: DatabricksAPI) -> None:
    paths, dirs = get_upload_paths_and_dirs(prod_workspace_path)
    upload_notebooks_and_folders(paths, dirs, handler)


if __name__ == "__main__":
    if len(sys.argv) < 2:
        raise RuntimeError("Expected Databricks token to passed as a command line argument.")

    else:
        config = read_config()
        databricks_handler = DatabricksAPI(host=config["host"], token=sys.argv[1])
        prod_workspace = config["production"]["workspace_path"]

        backup_path = f"{prod_workspace}-backup-{uuid1()}"
        copy_workspace_in_databricks(prod_workspace, backup_path, databricks_handler)
        try:
            delete_workspace(prod_workspace, databricks_handler)
            upload_from_repo(prod_workspace, databricks_handler)
        except [HTTPError, FileNotFoundError]:
            restore_production_workspace(backup_path, prod_workspace, databricks_handler)
        finally:
            delete_workspace(backup_path, databricks_handler)



from pathlib import Path
from typing import Dict

from jinja2 import Template

from forge_template.util.file_util import read_file, write_file


def render(content: Dict[str, str], template_path: Path, output_path: Path):
    template = Template(read_file(template_path))
    rendered = template.render(**content)
    write_file(output_path, rendered)
